"""Utility functions for file and paths solver."""

from collections.abc import Sequence
from pathlib import Path
from typing import Any, Literal

from numpy import ndarray
from numpy.typing import NDArray

from careamics.config.support import SupportedData
from careamics.dataset.dataset_utils import list_files, validate_source_target_files
from careamics.dataset_ng.image_stack_loader.zarr_utils import is_valid_uri

ItemType = Path | str | NDArray[Any]
"""Type of input items passed to the dataset."""

InputType = ItemType | Sequence[ItemType] | None
"""Type of input data passed to the dataset."""


def list_files_in_directory(
    data_type: Literal["tiff", "zarr", "czi", "custom"],
    input_data,
    target_data=None,
    extension_filter: str = "",
) -> tuple[list[Path], list[Path] | None]:
    """List files from input and target directories.

    Parameters
    ----------
    data_type : Literal["tiff", "zarr", "czi", "custom"]
        The type of data to validate.
    input_data : InputType
        Input data, can be a path to a folder, a list of paths, or a numpy array.
    target_data : Optional[InputType]
        Target data, can be None, a path to a folder, a list of paths, or a numpy
        array.
    extension_filter : str, default=""
        File extension filter to apply when listing files.

    Returns
    -------
    list[Path]
        A list of file paths for input data.
    list[Path] | None
        A list of file paths for target data, or None if target_data is None.
    """
    input_data = Path(input_data)

    # list_files will return a list with a single element if the path is a file with
    # the correct extension
    input_files = list_files(input_data, data_type, extension_filter)
    if target_data is None:
        return input_files, None
    else:
        target_data = Path(target_data)
        target_files = list_files(target_data, data_type, extension_filter)
        validate_source_target_files(input_files, target_files)
        return input_files, target_files


def convert_paths_to_pathlib(
    input_data: Sequence[str | Path],
    target_data: Sequence[str | Path] | None = None,
) -> tuple[list[Path], list[Path] | None]:
    """Create a list of file paths from the input and target data.

    Parameters
    ----------
    input_data : Sequence[str | Path]
        Input data, can be a path to a folder, or a list of paths.
    target_data : Sequence[str | Path] | None
        Target data, can be None, a path to a folder, or a list of paths.

    Returns
    -------
    list[Path]
        A list of file paths for input data.
    list[Path] | None
        A list of file paths for target data, or None if target_data is None.
    """
    input_files = [Path(item) if isinstance(item, str) else item for item in input_data]
    if target_data is None:
        return input_files, None
    else:
        target_files = [
            Path(item) if isinstance(item, str) else item for item in target_data
        ]
        validate_source_target_files(input_files, target_files)
        return input_files, target_files


def validate_input_target_type_consistency(
    input_data: InputType,
    target_data: InputType | None,
) -> None:
    """Validate if the input and target data types are consistent.

    Parameters
    ----------
    input_data : InputType
        Input data, can be a path to a folder, a list of paths, or a numpy array.
    target_data : Optional[InputType]
        Target data, can be None, a path to a folder, a list of paths, or a numpy
        array.

    Raises
    ------
    ValueError
        If the input and target data types are not consistent.
    """
    if input_data is not None and target_data is not None:
        if not isinstance(input_data, type(target_data)):
            raise ValueError(
                f"Inputs for input and target must be of the same type or None. "
                f"Got {type(input_data)} and {type(target_data)}."
            )
    if isinstance(input_data, list) and isinstance(target_data, list):
        if len(input_data) != len(target_data):
            raise ValueError(
                f"Inputs and targets must have the same length. "
                f"Got {len(input_data)} and {len(target_data)}."
            )
        if not isinstance(input_data[0], type(target_data[0])):
            raise ValueError(
                f"Inputs and targets must have the same type. "
                f"Got {type(input_data[0])} and {type(target_data[0])}."
            )


def validate_array_input(
    input_data: NDArray | list[NDArray],
    target_data: NDArray | list[NDArray] | None,
) -> tuple[list[NDArray], list[NDArray] | None]:
    """Validate if the input data is a numpy array.

    Parameters
    ----------
    input_data : InputType
        Input data, can be a path to a folder, a list of paths, or a numpy array.
    target_data : Optional[InputType]
        Target data, can be None, a path to a folder, a list of paths, or a numpy
        array.

    Returns
    -------
    list[numpy.ndarray]
        Validated input data.
    list[numpy.ndarray] | None
        Validated target data, None if the target data is None.

    Raises
    ------
    ValueError
        If the input data is not a numpy array or a list of numpy arrays.
    """
    if isinstance(input_data, ndarray):
        input_list = [input_data]

        if target_data is not None and not isinstance(target_data, ndarray):
            raise ValueError(
                f"Wrong target type. Expected numpy.ndarray, got {type(target_data)}. "
                f"Check the data_type parameter or your inputs."
            )
        target_list = [target_data] if target_data is not None else None
        return input_list, target_list
    elif isinstance(input_data, list):
        # TODO warn if wrong types inside list
        input_list = [array for array in input_data if isinstance(array, ndarray)]

        if target_data is None:
            target_list = None
        else:
            assert isinstance(target_data, list)
            target_list = [array for array in target_data if isinstance(array, ndarray)]
        return input_list, target_list
    else:
        raise ValueError(
            f"Wrong input type. Expected numpy.ndarray or list of numpy.ndarray, got "
            f"{type(input_data)}. Check the data_type parameter or your inputs."
        )


def validate_path_input(
    data_type: Literal["tiff", "zarr", "czi", "custom"],
    input_data: str | Path | list[str | Path],
    target_data: str | Path | list[str | Path] | None,
    extension_filter: str = "",
) -> tuple[list[Path], list[Path] | None]:
    """Validate if the input data is a path or a list of paths.

    Parameters
    ----------
    data_type : Literal["tiff", "zarr", "czi", "custom"]
        The type of data to validate.
    input_data : str | Path | list[str | Path]
        Input data, can be a path to a folder, a list of paths, or a numpy array.
    target_data : str | Path | list[str | Path] | None
        Target data, can be None, a path to a folder, a list of paths, or a numpy
        array.
    extension_filter : str, default=""
        File extension filter to apply when listing files.

    Returns
    -------
    list[Path]
        A list of file paths for input data.
    list[Path] | None
        A list of file paths for target data, or None if target_data is None.

    Raises
    ------
    ValueError
        If the input data is not a path or a list of paths.
    """
    if isinstance(input_data, (str, Path)):
        input_list, target_list = list_files_in_directory(
            data_type, input_data, target_data, extension_filter
        )
        return input_list, target_list
    elif isinstance(input_data, list):
        # TODO warn if wrong types inside list
        input_list = [
            Path(item)
            for item in input_data
            if isinstance(item, (str, Path)) and Path(item).exists()
        ]

        target_list = None
        if target_data is not None:
            assert isinstance(target_data, list)
            target_list = [
                Path(item)
                for item in target_data
                if isinstance(item, (str, Path)) and Path(item).exists()
            ]  # consistency with input is enforced by convert_paths_to_pathlib

        return convert_paths_to_pathlib(input_list, target_list)
    else:
        raise ValueError(
            f"Wrong input type, expected str or Path or list[str | Path], got "
            f"{type(input_data)}. Check the data_type parameter or your inputs."
        )


def validate_zarr_input(
    input_data: str | Path | list[str | Path],
    target_data: str | Path | list[str | Path] | None,
) -> tuple[list[str] | list[Path], list[str] | list[Path] | None]:
    """Validate if the input data corresponds a zarr input.

    Parameters
    ----------
    input_data : str | Path | list[str | Path]
        Input data, can be a path to a folder, to zarr file, a URI pointing to a zarr
        dataset, or a list.
    target_data : str | Path | list[str | Path] | None
        Target data, can be None.

    Returns
    -------
    list[str] or list[Path]
        A list of zarr URIs or path for input data.
    list[str] or list[Path] | None
        A list of zarr URIs or paths for target data, or None if target_data is None.

    Raises
    ------
    ValueError
        If the input and target data types are not consistent.
    ValueError
        If the input data is not a zarr URI or path, or a list of zarr URIs or paths.
    """
    # validate_input_target_type_consistency is called beforehand, ensuring the types
    # of input and target are the same
    if isinstance(input_data, (str, Path)):
        if Path(input_data).exists():
            # either a path to a folder or a zarr file
            # path to a folder will trigger collection of all zarr files in that folder
            assert target_data is None or isinstance(target_data, (str, Path))
            if target_data is not None and not Path(target_data).exists():
                raise ValueError(
                    f"Target provided as path, but does not exist: {target_data}."
                )

            return validate_path_input("zarr", input_data, target_data)
        elif isinstance(input_data, str) and is_valid_uri(input_data):
            input_list = [input_data]

            assert target_data is None or isinstance(target_data, str)
            if target_data is not None and not is_valid_uri(target_data):
                raise ValueError(
                    f"Wrong target type for zarr data. Expected a zarr URI, got "
                    f"{type(target_data)}."
                )
            target_list = [target_data] if target_data is not None else None
            return input_list, target_list
        else:
            raise ValueError(
                f"Wrong input type for zarr data. Expected a file URI or a path to a "
                f" file, got {input_data}. Path may not exist."
            )
    elif isinstance(input_data, list):
        # use first element as determinant of type
        if isinstance(input_data[0], (str, Path)):
            if Path(input_data[0]).exists():
                return validate_path_input("zarr", input_data, target_data)
            else:
                final_input_list = [
                    str(item) for item in input_data if is_valid_uri(item)
                ]
                if target_data is not None:
                    assert isinstance(target_data, list)
                    final_target_list = [
                        str(item) for item in target_data if is_valid_uri(item)
                    ]
                else:
                    final_target_list = None
                return final_input_list, final_target_list
        else:
            raise ValueError(
                f"Wrong input type for zarr data. Expected a list of file URIs or "
                f" paths to files, got {type(input_data[0])}."
            )
    else:
        raise ValueError(
            f"Wrong input type for zarr data. Expected a file URI, a path to a file, "
            f" or a list of those, got {type(input_data)}."
        )


def initialize_data_pair(
    data_type: Literal["array", "tiff", "zarr", "czi", "custom"],
    input_data: InputType,
    target_data: InputType | None = None,
    extension_filter: str = "",
    custom_loader: bool = False,
) -> tuple[InputType | list[InputType], InputType | list[InputType] | None]:
    """
    Initialize a pair of input and target data.

    Parameters
    ----------
    data_type : Literal["array", "tiff", "zarr", "czi", "custom"]
        The type of data to initialize.
    input_data : InputType
        Input data, can be None, a path to a folder, a list of paths, or a numpy
        array.
    target_data : InputType | None
        Target data, can be None, a path to a folder, a list of paths, or a numpy
        array.
    extension_filter : str, default=""
        File extension filter to apply when listing files.
    custom_loader : bool, default=False
        Whether a custom image stack loader is used.

    Returns
    -------
    list[numpy.ndarray] | list[pathlib.Path]
        Initialized input data. For file paths, returns a list of Path objects. For
        numpy arrays, returns the arrays directly.
    list[numpy.ndarray] | list[pathlib.Path] | None
        Initialized target data. For file paths, returns a list of Path objects. For
        numpy arrays, returns the arrays directly. Returns None if target_data is None.
    """
    if input_data is None:
        return None, None

    validate_input_target_type_consistency(input_data, target_data)

    if data_type == SupportedData.ARRAY:
        return validate_array_input(input_data, target_data)
    elif data_type in (SupportedData.TIFF, SupportedData.CZI):
        assert data_type != SupportedData.ARRAY.value  # for mypy

        if isinstance(input_data, (str, Path)):
            assert target_data is None or isinstance(target_data, (str, Path))

            return validate_path_input(data_type, input_data, target_data)
        elif isinstance(input_data, list):
            assert target_data is None or isinstance(target_data, list)

            return validate_path_input(data_type, input_data, target_data)
        else:
            raise ValueError(
                f"Unsupported input type for {data_type}: {type(input_data)}"
            )
    elif data_type == SupportedData.ZARR:
        return validate_zarr_input(input_data, target_data)
    elif data_type == SupportedData.CUSTOM:
        if custom_loader:
            return input_data, target_data
        return validate_path_input(data_type, input_data, target_data, extension_filter)
    else:
        raise NotImplementedError(f"Unsupported data type: {data_type}")



"""
Here, we have multiple folders, each containing images of a single channel.
"""

from collections import defaultdict
from functools import cache

import numpy as np

from .types import DataSplitType


def l2(x):
    return np.sqrt(np.mean(np.array(x) ** 2))


class MultiCropDset:
    def __init__(
        self,
        data_config,
        fpath: str,
        load_data_fn=None,
        val_fraction=None,
        test_fraction=None,
    ):

        assert (
            data_config.input_is_sum == True
        ), "This dataset is designed for sum of images"

        self._img_sz = data_config.image_size
        self._enable_rotation = data_config.enable_rotation_aug

        self._background_values = data_config.background_values
        self._data = load_data_fn(
            data_config, fpath, data_config.datasplit_type, val_fraction, test_fraction
        )

        # remove upper quantiles, crucial for removing puncta
        self.max_val = data_config.max_val
        if self.max_val is not None:
            for ch_idx, data in enumerate(self._data):
                if self.max_val[ch_idx] is not None:
                    for idx in range(len(data)):
                        data[idx][data[idx] > self.max_val[ch_idx]] = self.max_val[
                            ch_idx
                        ]

        # remove background values
        if self._background_values is not None:
            final_data_arr = []
            for ch_idx, data in enumerate(self._data):
                data_float = [x.astype(np.float32) for x in data]
                final_data_arr.append(
                    [x - self._background_values[ch_idx] for x in data_float]
                )
            self._data = final_data_arr

        print(
            f"{self.__class__.__name__} N:{len(self)} Rot:{self._enable_rotation} Ch:{len(self._data)} MaxVal:{self.max_val} Bg:{self._background_values}"
        )

    def get_max_val(self):
        return self.max_val

    def compute_mean_std(self):
        mean_tar_dict = defaultdict(list)
        std_tar_dict = defaultdict(list)
        mean_inp = []
        std_inp = []
        for _ in range(30000):
            crops = []
            for ch_idx in range(len(self._data)):
                crop = self.sample_crop(ch_idx)
                mean_tar_dict[ch_idx].append(np.mean(crop))
                std_tar_dict[ch_idx].append(np.std(crop))
                crops.append(crop)

            inp = 0
            for img in crops:
                inp += img

            mean_inp.append(np.mean(inp))
            std_inp.append(np.std(inp))

        output_mean = defaultdict(list)
        output_std = defaultdict(list)
        NC = len(self._data)
        for ch_idx in range(NC):
            output_mean["target"].append(np.mean(mean_tar_dict[ch_idx]))
            output_std["target"].append(l2(std_tar_dict[ch_idx]))

        output_mean["target"] = np.array(output_mean["target"]).reshape(NC, 1, 1)
        output_std["target"] = np.array(output_std["target"]).reshape(NC, 1, 1)

        output_mean["input"] = np.array([np.mean(mean_inp)]).reshape(1, 1, 1)
        output_std["input"] = np.array([l2(std_inp)]).reshape(1, 1, 1)
        return dict(output_mean), dict(output_std)

    def set_mean_std(self, mean_dict, std_dict):
        self._data_mean = mean_dict
        self._data_std = std_dict

    def get_mean_std(self):
        return self._data_mean, self._data_std

    def get_num_frames(self):
        return len(self._data)

    @cache
    def crop_probablities(self, ch_idx):
        sizes = np.array([np.prod(x.shape) for x in self._data[ch_idx]])
        return sizes / sizes.sum()

    def sample_crop(self, ch_idx):
        idx = None
        count = 0
        while idx is None:
            count += 1
            idx = np.random.choice(
                len(self._data[ch_idx]), p=self.crop_probablities(ch_idx)
            )
            data = self._data[ch_idx][idx]
            if data.shape[0] >= self._img_sz[0] and data.shape[1] >= self._img_sz[1]:
                h = np.random.randint(0, data.shape[0] - self._img_sz[0])
                w = np.random.randint(0, data.shape[1] - self._img_sz[1])
                return data[h : h + self._img_sz[0], w : w + self._img_sz[1]]
            elif count > 100:
                raise ValueError("Cannot find a valid crop")
            else:
                idx = None

        return None

    def len_per_channel(self, ch_idx):
        return np.sum([np.prod(x.shape) for x in self._data[ch_idx]]) / np.prod(
            self._img_sz
        )

    def imgs_for_patch(self):
        return [self.sample_crop(ch_idx) for ch_idx in range(len(self._data))]

    def __len__(self):
        len_per_channel = [
            self.len_per_channel(ch_idx) for ch_idx in range(len(self._data))
        ]
        return int(np.max(len_per_channel))

    def _rotate(self, img_tuples):
        return self._rotate2D(img_tuples)

    def _rotate2D(self, img_tuples):
        img_kwargs = {}
        for i, img in enumerate(img_tuples):
            for k in range(len(img)):
                img_kwargs[f"img{i}_{k}"] = img[k]

        keys = list(img_kwargs.keys())
        self._rotation_transform.add_targets({k: "image" for k in keys})
        rot_dic = self._rotation_transform(image=img_tuples[0][0], **img_kwargs)

        rotated_img_tuples = []
        for i, img in enumerate(img_tuples):
            if len(img) == 1:
                rotated_img_tuples.append(rot_dic[f"img{i}_0"][None])
            else:
                rotated_img_tuples.append(
                    np.concatenate(
                        [rot_dic[f"img{i}_{k}"][None] for k in range(len(img))], axis=0
                    )
                )

        return rotated_img_tuples

    def _compute_input(self, imgs):
        inp = 0
        for img in imgs:
            inp += img

        inp = inp[None]
        inp = (inp - self._data_mean["input"]) / (self._data_std["input"])
        return inp

    def _compute_target(self, imgs):
        imgs = np.stack(imgs)
        target = (imgs - self._data_mean["target"]) / (self._data_std["target"])
        return target

    def __getitem__(self, idx):
        imgs = self.imgs_for_patch()
        if self._enable_rotation:
            imgs = self._rotate(imgs)

        inp = self._compute_input(imgs)
        target = self._compute_target(imgs)
        return inp, target



"""Configuration for N2V."""

from typing import Self

import numpy as np
from pydantic import model_validator

from careamics.config.algorithms import N2VAlgorithm
from careamics.config.data.patching_strategies import RandomPatchingConfig

from .ng_configuration import NGConfiguration


class N2VConfiguration(NGConfiguration):
    """N2V-specific configuration."""

    algorithm_config: N2VAlgorithm

    @model_validator(mode="after")
    def validate_n2v_mask_pixel_perc(self: Self) -> Self:
        """
        Validate that there will always be at least one blind-spot pixel in every patch.

        The probability of creating a blind-spot pixel is a function of the chosen
        masked pixel percentage and patch size.

        Returns
        -------
        Self
            Validated configuration.

        Raises
        ------
        ValueError
            If the probability of masking a pixel within a patch is less than 1 for the
            chosen masked pixel percentage and patch size.
        """
        if self.data_config.mode == "training":
            assert isinstance(self.data_config.patching, RandomPatchingConfig)

            mask_pixel_perc = self.algorithm_config.n2v_config.masked_pixel_percentage
            patch_size = self.data_config.patching.patch_size
            expected_area_per_pixel = 1 / (mask_pixel_perc / 100)

            n_dims = 3 if self.algorithm_config.model.is_3D() else 2
            patch_size_lower_bound = int(
                np.ceil(expected_area_per_pixel ** (1 / n_dims))
            )
            required_patch_size = tuple(
                2 ** int(np.ceil(np.log2(patch_size_lower_bound)))
                for _ in range(n_dims)
            )
            required_mask_pixel_perc = (1 / np.prod(patch_size)) * 100

            if expected_area_per_pixel > np.prod(patch_size):
                raise ValueError(
                    "The probability of creating a blind-spot pixel within a patch is "
                    f"below 1, for a patch size of {patch_size} with a masked pixel "
                    f"percentage of {mask_pixel_perc}%. Either increase the patch size "
                    f"to {required_patch_size} or increase the masked pixel percentage "
                    f"to at least {required_mask_pixel_perc}%."
                )

        return self



from typing import Any, Union

from pydantic import BaseModel, ConfigDict

from .types import DataSplitType, DataType, TilingMode


# TODO: check if any bool logic can be removed
class MicroSplitDataConfig(BaseModel):
    model_config = ConfigDict(validate_assignment=True, extra="allow")

    data_type: Union[DataType, str] | None  # TODO remove or refactor!!
    """Type of the dataset, should be one of DataType"""

    depth3D: int | None = 1
    """Number of slices in 3D. If data is 2D depth3D is equal to 1"""

    datasplit_type: DataSplitType | None = None
    """Whether to return training, validation or test split, should be one of
    DataSplitType"""

    num_channels: int | None = 2
    """Number of channels in the input"""

    # TODO: remove ch*_fname parameters, should be parsed automatically from a name list
    ch1_fname: str | None = None
    ch2_fname: str | None = None
    ch_input_fname: str | None = None

    input_is_sum: bool | None = False
    """Whether the input is the sum or average of channels"""

    input_idx: int | None = None
    """Index of the channel where the input is stored in the data"""

    target_idx_list: list[int] | None = None
    """Indices of the channels where the targets are stored in the data"""

    # TODO: where are there used?
    start_alpha: Any | None = None
    end_alpha: Any | None = None

    image_size: tuple  # TODO: revisit, new model_config uses tuple
    """Size of one patch of data"""

    grid_size: Union[int, tuple[int, int, int]] | None = None
    """Frame is divided into square grids of this size. A patch centered on a grid 
    having size `image_size` is returned. Grid size not used in training,
    used only during val / test, grid size controls the overlap of the patches"""

    empty_patch_replacement_enabled: bool | None = False
    """Whether to replace the content of one of the channels
    with background with given probability"""
    empty_patch_replacement_channel_idx: Any | None = None
    empty_patch_replacement_probab: Any | None = None
    empty_patch_max_val_threshold: Any | None = None

    uncorrelated_channels: bool | None = False
    """Replace the content in one of the channels with given probability to make
    channel content 'uncorrelated'"""
    uncorrelated_channel_probab: float | None = 0.5

    poisson_noise_factor: float | None = -1
    """The added poisson noise factor"""

    synthetic_gaussian_scale: float | None = 0.1

    # TODO: set to True in training code, recheck
    input_has_dependant_noise: bool | None = False

    # TODO: sometimes max_val differs between runs with fixed seeds with noise enabled
    enable_gaussian_noise: bool | None = False
    """Whether to enable gaussian noise"""

    # TODO: is this parameter used?
    allow_generation: bool = False

    # TODO: both used in IndexSwitcher, insure correct passing
    training_validtarget_fraction: Any = None
    deterministic_grid: Any = None

    # TODO: why is this not used?
    enable_rotation_aug: bool | None = False

    max_val: Union[float, tuple] | None = None
    """Maximum data in the dataset. Is calculated for train split, and should be
    externally set for val and test splits."""

    overlapping_padding_kwargs: Any = None
    """Parameters for np.pad method"""

    # TODO: remove this parameter, controls debug print
    print_vars: bool | None = False

    # Hard-coded parameters (used to be in the config file)
    normalized_input: bool = True
    """If this is set to true, then one mean and stdev is used
                for both channels. Otherwise, two different mean and stdev are used."""
    use_one_mu_std: bool | None = True

    # TODO: is this parameter used?
    train_aug_rotate: bool | None = False
    enable_random_cropping: bool | None = True

    multiscale_lowres_count: int | None = None
    """Number of LC scales"""

    tiling_mode: TilingMode | None = TilingMode.ShiftBoundary

    target_separate_normalization: bool | None = True

    mode_3D: bool | None = False
    """If training in 3D mode or not"""

    trainig_datausage_fraction: float | None = 1.0

    validtarget_random_fraction: float | None = None

    validation_datausage_fraction: float | None = 1.0

    random_flip_z_3D: bool | None = False

    padding_kwargs: dict = {"mode": "reflect"}  # TODO remove !!

    def __init__(self, **data):
        # Convert string data_type to enum if needed
        if "data_type" in data and isinstance(data["data_type"], str):
            try:
                data["data_type"] = DataType[data["data_type"]]
            except KeyError:
                # Keep original value to let validation handle the error
                pass
        super().__init__(**data)

    # TODO add validators !



"""Algorithms supported by CAREamics."""

from __future__ import annotations

from careamics.utils import BaseEnum


class SupportedAlgorithm(str, BaseEnum):
    """Algorithms available in CAREamics.

    These definitions are the same as the keyword `name` of the algorithm
    configurations.
    """

    N2V = "n2v"
    """Noise2Void algorithm, a self-supervised approach based on blind denoising."""

    CARE = "care"
    """Content-aware image restoration, a supervised algorithm used for a variety
    of tasks."""

    N2N = "n2n"
    """Noise2Noise algorithm, a self-supervised denoising scheme based on comparing
    noisy images of the same sample."""

    MUSPLIT = "musplit"  # TODO remove
    """An image splitting approach based on ladder VAE architectures."""

    MICROSPLIT = "microsplit"
    """A micro-level image splitting approach based on ladder VAE architectures."""

    DENOISPLIT = "denoisplit"
    """An image splitting and denoising approach based on ladder VAE architectures."""

    HDN = "hdn"
    """Hierarchical Denoising Network, an unsupervised denoising algorithm"""

    PN2V = "pn2v"
    """Probabilistic Noise2Void. A extension of Noise2Void is not restricted to Gaussian
    noise models or Gaussian intensity predictions."""



"""Module for the `TilingStrategy` class."""

import itertools
from collections.abc import Sequence
from math import prod

from .patching_strategy_protocol import TileSpecs


class TilingStrategy:
    """
    The tiling strategy should be used for prediction. The `get_patch_specs`
    method returns `TileSpec` dictionaries that contains information on how to
    stitch the tiles back together to create the full image.
    """

    def __init__(
        self,
        data_shapes: Sequence[Sequence[int]],
        patch_size: Sequence[int],
        overlaps: Sequence[int],
    ):
        """
        The tiling strategy should be used for prediction. The `get_patch_specs`
        method returns `TileSpec` dictionaries that contains information on how to
        stitch the tiles back together to create the full image.

        Parameters
        ----------
        data_shapes : sequence of (sequence of int)
            The shapes of the underlying data. Each element is the dimension of the
            axes SC(Z)YX.
        patch_size : sequence of int
            The size of the tile. The sequence will have length 2 or 3, for 2D and 3D
            data respectively.
        overlaps : sequence of int
            How much a tile will overlap with adjacent tiles in each spatial dimension.
        """
        self.data_shapes = data_shapes
        self.patch_size = patch_size
        self.overlaps = overlaps
        # patch_size and overlap should have same length validated in pydantic configs
        self.tile_specs: list[TileSpecs] = self._generate_specs()

    @property
    def n_patches(self) -> int:
        """
        The number of patches that this patching strategy will return.

        It also determines the maximum index that can be given to `get_patch_spec`.
        """
        return len(self.tile_specs)

    def get_patch_spec(self, index: int) -> TileSpecs:
        """Return the tile specs for a given index.

        Parameters
        ----------
        index : int
            A patch index.

        Returns
        -------
        TileSpecs
            A dictionary that specifies a single patch in a series of `ImageStacks`.
        """
        return self.tile_specs[index]

    # Note: this is used by the FileIterSampler
    def get_patch_indices(self, data_idx: int) -> Sequence[int]:
        """
        Get the patch indices will return patches for a specific `image_stack`.

        The `image_stack` corresponds to the given `data_idx`.

        Parameters
        ----------
        data_idx : int
            An index that corresponds to a given `image_stack`.

        Returns
        -------
        sequence of int
            A sequence of patch indices, that when used to index the `CAREamicsDataset
            will return a patch that comes from the `image_stack` corresponding to the
            given `data_idx`.
        """
        return [
            i
            for i, patch_spec in enumerate(self.tile_specs)
            if patch_spec["data_idx"] == data_idx
        ]

    def _generate_specs(self) -> list[TileSpecs]:
        tile_specs: list[TileSpecs] = []
        for data_idx, data_shape in enumerate(self.data_shapes):
            spatial_shape = data_shape[2:]

            # spec info for each axis
            axis_specs: list[tuple[list[int], list[int], list[int], list[int]]] = [
                self._compute_1d_coords(
                    axis_size, self.patch_size[axis_idx], self.overlaps[axis_idx]
                )
                for axis_idx, axis_size in enumerate(spatial_shape)
            ]

            # combine by using zip
            all_coords, all_stitch_coords, all_crop_coords, all_crop_size = zip(
                *axis_specs, strict=False
            )

            # number of tiles for this data_idx
            n_tiles = prod(len(dim) for dim in all_coords) * data_shape[0]

            # patches will be the same for each sample in a stack
            for sample_idx in range(data_shape[0]):
                # iterate through all combinations using itertools.product
                for coords, stitch_coords, crop_coords, crop_size in zip(
                    itertools.product(*all_coords),
                    itertools.product(*all_stitch_coords),
                    itertools.product(*all_crop_coords),
                    itertools.product(*all_crop_size),
                    strict=False,
                ):
                    tile_specs.append(
                        {
                            # PatchSpecs
                            "data_idx": data_idx,
                            "sample_idx": sample_idx,
                            "coords": coords,
                            "patch_size": self.patch_size,
                            # TileSpecs additional fields
                            "crop_coords": crop_coords,
                            "crop_size": crop_size,
                            "stitch_coords": stitch_coords,
                            "total_tiles": n_tiles,
                        }
                    )

        return tile_specs

    @staticmethod
    def _compute_1d_coords(
        axis_size: int, patch_size: int, overlap: int
    ) -> tuple[list[int], list[int], list[int], list[int]]:
        """
        Computes the TileSpec information for a single axis.

        Parameters
        ----------
        axis_size : int
            The size of the axis.
        patch_size : int
            The tile size.
        overlap : int
            The tile overlap.

        Returns
        -------
        coords: list of int
            The top-left (and first z-slice for 3D data) of a tile, in coords relative
            to the image.
        stitch_coords: list of int
            Where the tile will be stitched back into an image, taking into account
            that the tile will be cropped, in coords relative to the image.
        crop_coords: list of int
            The top-left side of where the tile will be cropped, in coordinates relative
            to the tile.
        crop_size: list of int
            The size of the cropped tile.
        """
        coords: list[int] = []
        stitch_coords: list[int] = []
        crop_coords: list[int] = []
        crop_size: list[int] = []

        step = patch_size - overlap
        for i in range(0, max(1, axis_size - overlap), step):
            if i == 0:
                coords.append(i)
                crop_coords.append(0)
                stitch_coords.append(0)
                if axis_size <= patch_size:
                    crop_size.append(axis_size)
                else:
                    crop_size.append(patch_size - overlap // 2)
            elif (0 < i) and (i + patch_size < axis_size):
                coords.append(i)
                crop_coords.append(overlap // 2)
                stitch_coords.append(coords[-1] + crop_coords[-1])
                crop_size.append(patch_size - overlap)
            else:
                previous_crop_size = crop_size[-1] if crop_size else 1
                previous_stitch_coord = stitch_coords[-1] if stitch_coords else 0
                previous_tile_end = previous_stitch_coord + previous_crop_size

                coords.append(max(0, axis_size - patch_size))
                stitch_coords.append(previous_tile_end)
                crop_coords.append(stitch_coords[-1] - coords[-1])
                crop_size.append(axis_size - stitch_coords[-1])

        return (
            coords,
            stitch_coords,
            crop_coords,
            crop_size,
        )



from pytorch_lightning.callbacks import EarlyStopping, ModelCheckpoint

from careamics.config.lightning.callbacks.callback_config import (
    CheckpointConfig,
    EarlyStoppingConfig,
)


def test_defaults_checkpoint():
    """Test that a default CheckpointConfig can initialize a ModelCheckpoint
    callback."""
    config = CheckpointConfig()
    ModelCheckpoint(**config.model_dump())


def test_defaults_early_stopping():
    """Test that a default EarlyStoppingConfig can initialize an EarlyStopping
    callback."""
    config = EarlyStoppingConfig()
    EarlyStopping(**config.model_dump())



"""Module containing different strategies for writing predictions."""

from pathlib import Path
from typing import Protocol

from careamics.dataset_ng.dataset import ImageRegionData


class WriteStrategy(Protocol):
    """Protocol for write strategy classes."""

    def write_batch(
        self,
        dirpath: Path,
        predictions: list[ImageRegionData],
    ) -> None:
        """
        WriteStrategy subclasses must contain this function to write a batch.

        Parameters
        ----------
        dirpath : Path
            Path to directory to save predictions to.
        predictions : list[ImageRegionData]
            Decollated predictions.
        """
        ...



"""Deep-learning model configurations."""

__all__ = ["ArchitectureConfig", "LVAEConfig", "UNetConfig"]

from .architecture_config import ArchitectureConfig
from .lvae_config import LVAEConfig
from .unet_config import UNetConfig



"""Algorithm configurations."""

__all__ = [
    "CAREAlgorithm",
    "HDNAlgorithm",
    "MicroSplitAlgorithm",
    "N2NAlgorithm",
    "N2VAlgorithm",
    "PN2VAlgorithm",
    "UNetBasedAlgorithm",
    "VAEBasedAlgorithm",
]

from .care_algorithm_config import CAREAlgorithm
from .hdn_algorithm_config import HDNAlgorithm
from .microsplit_algorithm_config import MicroSplitAlgorithm
from .n2n_algorithm_config import N2NAlgorithm
from .n2v_algorithm_config import N2VAlgorithm
from .pn2v_algorithm_config import PN2VAlgorithm
from .unet_algorithm_config import UNetBasedAlgorithm
from .vae_algorithm_config import VAEBasedAlgorithm



from __future__ import annotations

import re
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

import numpy as np
from numpy.typing import NDArray

try:
    from pylibCZIrw.czi import CziReader, Rectangle, open_czi

    pyczi_available = True
except ImportError:
    pyczi_available = False

if TYPE_CHECKING:
    try:
        from pylibCZIrw.czi import CziReader, Rectangle, open_czi
    except ImportError:
        CziReader = Rectangle = open_czi = None  # type: ignore


class CziImageStack:
    """
    A class for extracting patches from an image stack that is stored as a CZI file.

    Parameters
    ----------
    data_path : str or Path
        Path to the CZI file.

    scene : int, optional
        Index of the scene to extract.

        A single CZI file can contain multiple "scenes", which are stored alongside each
        other at different coordinates in the image plane, often separated by empty
        space. Specifying this argument will read only the single scene with that index
        from the file. Think of it as cropping the CZI file to the region where that
        scene is located.

        If no scene index is specified, the entire image will be read. In case it
        contains multiple scenes, they will all be present in the resulting image.
        This is usually not desirable due to the empty space between them.
        In general, only omit this argument or set it to `None` if you know that
        your CZI file does not contain any scenes.

        The static function :py:meth:`get_bounding_rectangles` can be used to find out
        how many scenes a given file contains and what their bounding rectangles are.

        The scene can also be provided as part of `data_path` by appending an `"@"`
        followed by the scene index to the filename.

    depth_axis : {"none", "Z", "T"}, default: "none"
        Which axis to use as depth-axis for providing 3-D patches.

        - `"none"`: Only provide 2-D patches. If a Z or T dimension is present in the
          data, they will be combined into the sample dimension `S`.
        - `"Z"`: Use the Z-axis as depth-axis. If a T axis is present as well, it will
          be merged into the sample dimensions `S`.
        - `"T"`: Use the T-axis as depth-axis. If a Z axis is present as well, it will
          be merged into the sample dimensions `S`.

    Attributes
    ----------
    source : Path
        Path to the CZI file, including the scene index if specified.
    data_path : Path
        Path to the CZI file without scene index.
    scene : int or None
        Index of the scene to extract, or None if not specified.
    data_shape : Sequence[int]
        The shape of the data in the order `(SC(Z)YX)`.
    axes : str
        The axes in the CZI file corresponding to the dimensions in `data_shape`.
        The following values can occur:

        - "SCZYX" for 3-D volumes if `depth_axis` is `"Z"`.
        - "SCTYX" for time-series if `depth_axis` is `"T"`.
        - "SCYX" if `depth_axis` is `"none"`.

        The axis `S` (sample) is the only one not mapping one-to-one to an axis in the
        CZI file but combines all remaining axes present in the file into one.

    Examples
    --------
    Create an image stack for the first scene in a CZI file:
    >>> stack = CziImageStack("path/to/file.czi", scene=0)  # doctest: +SKIP

    Alternatively, the scene index can also be provided as part of the filename.
    This is mainly intended for re-creating an image stack from the `source` property:
    >>> stack = CziImageStack("path/to/file.czi@0")  # doctest: +SKIP
    >>> stack2 = CziImageStack(stack.source)  # doctest: +SKIP

    If the CZI file contains a third dimension (Z or T) and you want to perform 3-D
    denoising, you need to explicitly set `depth_axis` to `"Z"` or `"T"`:
    >>> stack_2d = CziImageStack("path/to/file.czi", scene=0)  # doctest: +SKIP
    >>> stack_2d.axes, stack_2d.data_shape  # doctest: +SKIP
    ('SCYX', [40, 1, 512, 512])
    >>> stack_3d = CziImageStack(  # doctest: +SKIP
    ...     "path/to/file.czi", scene=0, depth_axis="Z"
    ... )
    >>> stack_3d.axes, stack_3d.data_shape  # doctest: +SKIP
    ('SCZYX', [4, 1, 10, 512, 512])
    """

    def __init__(
        self,
        data_path: str | Path,
        scene: int | None = None,
        depth_axis: Literal["none", "Z", "T"] = "none",
    ) -> None:
        if not pyczi_available:
            raise ImportError(
                "The CZI image stack requires the `pylibCZIrw` package to be installed."
                " Please install it with `pip install careamics[czi]`."
            )

        _data_path = Path(data_path)

        # Check for scene encoded in filename.
        # Normally, file path and scene should be provided as separate arguments but
        # we would also like to support using the `source` property to re-create the
        # CZI image stack. In this case, the scene index is encoded in the file path.
        scene_matches = re.match(r"^(.*)@(\d+)$", _data_path.name)
        if scene_matches:
            if scene is not None:
                raise ValueError(
                    f"Scene index is specified in the filename ({_data_path.name}) and "
                    f"as an argument ({scene}). Please specify only one."
                )
            _data_path = _data_path.parent / scene_matches.group(1)
            scene = int(scene_matches.group(2))

        # Set variables
        self.data_path = _data_path
        self.scene = scene
        self._depth_axis = depth_axis

        # Open CZI file
        self._czi = CziReader(str(self.data_path))

        # Determine metadata
        self.axes, self.data_shape, self._bounding_rectangle, self._sample_axes = (
            self._get_shape()
        )
        self.data_dtype = np.float32

    def __del__(self):
        if hasattr(self, "_czi"):
            # Close CZI file
            self._czi.close()

    def __getstate__(self) -> dict[str, Any]:
        # Remove CziReader object from state to avoid pickling issues
        state = self.__dict__.copy()
        del state["_czi"]
        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        # Reopen CZI file after unpickling
        self.__dict__.update(state)
        self._czi = CziReader(str(self.data_path))

    # TODO: we append the scene index to the file name
    #       - not sure if this is a good approach
    @property
    def source(self) -> Path:
        filename = self.data_path.name
        if self.scene is not None:
            filename = f"{filename}@{self.scene}"
        return self.data_path.parent / filename

    def extract_patch(
        self,
        sample_idx: int,
        channels: Sequence[int] | None,  # `channels = None` to select all channels
        coords: Sequence[int],
        patch_size: Sequence[int],
    ) -> NDArray:
        # check that channels are within bounds
        if channels is not None:
            max_channel = self.data_shape[1] - 1  # channel is second dimension
            for ch in channels:
                if ch > max_channel:
                    raise ValueError(
                        f"Channel index {ch} is out of bounds for data with "
                        f"{self.data_shape[1]} channels. Check the provided `channels` "
                        f"parameter in the configuration for erroneous channel "
                        f"indices."
                    )

        # Determine 3rd dimension (T, Z or none)
        if len(coords) == 3:
            if len(self.axes) != 5:
                raise ValueError(
                    f"Requested a 3D patch from a 2D image stack with axes {self.axes}."
                )
            third_dim = self.axes[2]
            third_dim_offset, third_dim_size = coords[0], patch_size[0]
        else:
            if len(self.axes) != 4:
                raise ValueError(
                    f"Requested a 2D patch from a 3D image stack with axes {self.axes}."
                )
            third_dim = None
            third_dim_offset, third_dim_size = 0, 1

        # Set up ROI to extract from each plane as (x, y, w, h)
        roi = (
            self._bounding_rectangle.x + coords[-1],
            self._bounding_rectangle.y + coords[-2],
            patch_size[-1],
            patch_size[-2],
        )

        # Create output array of shape (C, Z, Y, X)
        n_channels = self.data_shape[1] if channels is None else len(channels)
        patch = np.zeros(
            (n_channels, third_dim_size, *patch_size[-2:]), dtype=np.float32
        )

        # Set up plane to index `sample_idx`
        sample_shape = list(self._sample_axes.values())
        sample_indices = np.unravel_index(sample_idx, sample_shape)
        plane = {
            dimension: int(index)
            for dimension, index in zip(
                self._sample_axes.keys(), sample_indices, strict=False
            )
        }

        # Read XY planes sequentially
        channel_iter: Iterable
        if channels is None:
            channel_iter = range(self.data_shape[1])  # iter over number of requested C
        else:
            channel_iter = list(channels)

        # for each channel
        for patch_channel, data_channel in enumerate(channel_iter):
            # pull plane with the given channel and 3rd dim index
            for third_dim_index in range(third_dim_size):
                plane["C"] = data_channel
                if third_dim is not None:
                    plane[third_dim] = third_dim_offset + third_dim_index

                    # check if in bounds
                    if 0 > plane[third_dim] or plane[third_dim] >= self.data_shape[-3]:
                        continue

                # read plane
                extracted_roi = self._czi.read(roi=roi, plane=plane, scene=self.scene)
                if extracted_roi.ndim == 3:
                    if extracted_roi.shape[-1] > 1:
                        raise ValueError(
                            "CZI files with RGB channels are currently not "
                            "supported."
                        )

                    # remove channel dimension
                    extracted_roi = extracted_roi.squeeze(-1)

                # add requested channel into the patch
                patch[patch_channel, third_dim_index] = extracted_roi

        # Remove dummy 3rd dimension for 2-D data
        if third_dim is None:
            patch = patch.squeeze(1)

        return patch

    def _get_shape(self) -> tuple[str, list[int], Rectangle, dict[str, int]]:
        """Determines the shape of the selected scene.

        Returns
        -------
        axes : str
            String specifying the axis order. Examples:

            - "SCZYX" for 3-D volumes if `depth_axis` is `"Z"`.
            - "SCTYX" for time-series if `depth_axis` is `"T"`.
            - "SCYX" if `depth_axis` is `"none"`.

            The axis `S` is the sample dimension and combines all remaining axes
            present in the data.

        shape : list[int]
            The size of each axis, in the order listed in `axes`.

        bounding_rectangle : Rectangle
            The bounding rectangle of the scene in pixels. The rectangle is
            defined by its top-left corner (x, y) and its width and height (w, h).

        sample_axes : dict[str, int]
            A dictionary with information about the remaining axes used for the
            sample dimension.
            The keys are the axis names (e.g., "T", "Z") and the values are their
            respective sizes.
        """
        # Get CZI dimensions
        total_bbox = self._czi.total_bounding_box_no_pyramid
        if self.scene is None:
            bounding_rectangle = self._czi.total_bounding_rectangle_no_pyramid
        else:
            bounding_rectangle = self._czi.scenes_bounding_rectangle_no_pyramid[
                self.scene
            ]

        # Determine if T and Z axis are present
        # Note: An axis of size 1 is as good as no axis since we cannot use it for 3-D
        # denoising.
        has_time = "T" in total_bbox and (total_bbox["T"][1] - total_bbox["T"][0]) > 1
        has_depth = "Z" in total_bbox and (total_bbox["Z"][1] - total_bbox["Z"][0]) > 1

        # Determine axis order depending on `depth_axis`
        if self._depth_axis == "Z":
            axes = "SCZYX"
            if not has_depth:
                raise RuntimeError(
                    f"The CZI file {self.data_path} does not contain a Z axis to use "
                    'for 3-D denoising. Consider setting `axes="YX"` or '
                    '`depth_axis="none"` to perform 2-D denoising instead.'
                )
        elif self._depth_axis == "T":
            axes = "SCTYX"
            if not has_time:
                raise RuntimeError(
                    f"The CZI file {self.data_path} does not contain a T axis to use "
                    'for 3-D denoising. Consider setting `axes="YX"` or '
                    '`depth_axis="none"` to perform 2-D denoising instead.'
                )
        else:
            axes = "SCYX"

        # Calculcate size of sample dimension S, combining all axes not used elsewhere.
        # This could, for example, be a time axis. If we only perform 2-D denoising, a
        # potentially present Z axis would also be used as sample dimension. If both,
        # T and Z, are present, both need to be combined into the sample dimension.
        # The same needs to be done to any other potentially present axis in the CZI
        # file which is not a spatial or channel axis.
        # The following code calculates the size of the combined sample axis.
        sample_axes = {}
        sample_size = 1
        for dimension, (start, end) in total_bbox.items():
            if dimension not in axes:
                sample_axes[dimension] = end - start
                sample_size *= end - start

        # Determine data shape
        shape = []
        for dimension in axes:
            if dimension == "S":
                shape.append(sample_size)
            elif dimension == "Y":
                shape.append(bounding_rectangle.h)
            elif dimension == "X":
                shape.append(bounding_rectangle.w)
            elif dimension in total_bbox:
                shape.append(total_bbox[dimension][1] - total_bbox[dimension][0])
            else:
                shape.append(1)

        return axes, shape, bounding_rectangle, sample_axes

    @classmethod
    def get_bounding_rectangles(
        cls, czi: Path | str | CziReader
    ) -> dict[int | None, Rectangle]:
        """Gets the bounding rectangles of all scenes in a CZI file.

        Parameters
        ----------
        czi : Path or str or pyczi.CziReader
            Path to the CZI file or an already opened file as CziReader object.

        Returns
        -------
        dict[int | None, Rectangle]
            A dictionary mapping scene indices to their bounding rectangles in the
            format `(x, y, w, h)`.
            If no scenes are present in the CZI file, the returned dictionary will
            have only one entry with key `None`, whose bounding rectangle covers the
            entire image.
        """
        if not isinstance(czi, CziReader):
            with open_czi(str(czi)) as czi_reader:
                return cls.get_bounding_rectangles(czi_reader)

        scenes_bounding_rectangle = czi.scenes_bounding_rectangle_no_pyramid
        if len(scenes_bounding_rectangle) >= 1:
            # Ensure keys are int | None for type compatibility
            return {int(k): v for k, v in scenes_bounding_rectangle.items()}
        else:
            return {None: czi.total_bounding_rectangle_no_pyramid}



__all__ = [
    "CziImageStack",
    "FileImageStack",
    "GenericImageStack",
    "ImageStack",
    "InMemoryImageStack",
    "ZarrImageStack",
]

from .czi_image_stack import CziImageStack
from .file_image_stack import FileImageStack
from .image_stack_protocol import GenericImageStack, ImageStack
from .in_memory_image_stack import InMemoryImageStack
from .zarr_image_stack import ZarrImageStack



"""MicroSplit patch synthesis."""

# --- PROOF OF PRINCIPLE ---


from collections.abc import Callable, Sequence
from typing import Any, Literal, NamedTuple

import numpy as np
from numpy.typing import NDArray

from .dataset import ImageRegionData
from .image_stack import ImageStack
from .patch_extractor import PatchExtractor
from .patch_filter import PatchFilterProtocol
from .patching_strategies import PatchingStrategy, PatchSpecs


# TODO: better name
# mirrors format of ImageRegionData
class UncorrelatedRegionData(NamedTuple):
    data: NDArray
    source: Sequence[str | Literal["array"]]
    data_shape: Sequence[Sequence[int]]
    dtype: Sequence[str]  # dtype should be str for collate
    axes: Sequence[str]
    region_spec: Sequence[PatchSpecs]


# --- for finding empty / signal channel patches in loop
def is_empty(filter: PatchFilterProtocol) -> Callable[[NDArray[Any]], bool]:
    def is_empty_check(patch: NDArray[Any]) -> bool:
        return filter.filter_out(patch)

    return is_empty_check


def is_not_empty(filter: PatchFilterProtocol) -> Callable[[NDArray[Any]], bool]:
    def is_not_empty_check(patch: NDArray[Any]) -> bool:
        return not filter.filter_out(patch)

    return is_not_empty_check


# ---


def create_default_input_target(
    idx: int,
    patch_extractor: PatchExtractor[ImageStack],
    patching_strategy: PatchingStrategy,
    alphas: list[float],
    axes: str,  # annoyingly have to supply this to image region
) -> tuple[ImageRegionData, ImageRegionData]:
    """
    Create a default MicroSplit patch with synthetically summed input.

    Parameters
    ----------
    idx: int
        The dataset index.
    patch_extractor: PatchExtractor
        Used to extract patches from the data.
    patching_strategy: PatchingStrategy
        Patch locations will be sampled using the patching strategy.
    alphas: list[float]
        Weights for each channel for creating the synthetic input with summation.
    axes: str
        The axes of the data. This is only used to populate metadata.

    Returns
    -------
    input_region: ImageRegionData
        The input patch and its metadata, the data has the dimension L(Z)YX.
    target_region: ImageRegionData
        The target patch and its metadata, the data has the dimensions C(Z)YX.
    """
    patch_spec = patching_strategy.get_patch_spec(idx)
    patches = extract_microsplit_patch(patch_extractor, patch_spec)

    ndims = len(patches.shape) - 1
    alpha_broadcast = np.array(alphas)[:, *(np.newaxis for _ in range(ndims))]
    # weight channels by alphas then sum on the channel axis
    # input dims will be L(Z)YX
    input_patch = (alpha_broadcast * patches).sum(axis=0)
    target_patch = patches[:, 0, ...]  # first L patch

    data_idx = patch_spec["data_idx"]
    input_region = ImageRegionData(
        input_patch,
        source=str(patch_extractor.image_stacks[data_idx].source),
        data_shape=patch_extractor.image_stacks[data_idx].data_shape,
        dtype=str(patch_extractor.image_stacks[data_idx].data_dtype),
        axes=axes,
        region_spec=patch_spec,
        additional_metadata={},
    )
    target_region = ImageRegionData(
        target_patch,
        source=str(patch_extractor.image_stacks[data_idx].source),
        data_shape=patch_extractor.image_stacks[data_idx].data_shape,
        dtype=str(patch_extractor.image_stacks[data_idx].data_dtype),
        axes=axes,
        region_spec=patch_spec,
        additional_metadata={},
    )
    return input_region, target_region


def create_uncorrelated_input_target(
    patches: NDArray[Any],
    patch_specs: list[PatchSpecs],
    alphas: list[float],
    patch_extractor: PatchExtractor[ImageStack],  # for metadata
    axes: str,  # mirroring imageregion
) -> tuple[UncorrelatedRegionData, UncorrelatedRegionData]:
    """
    Create MicroSplit target and synthetically summed input with metadata.

    Parameters
    ----------
    patches: NDArray
        Patches with dimensions LC(Z)YX, where L contains the lateral context at
        multiple scales.
    patch_specs: list[PatchSpecs]
        The patch specs for each channel.
    alphas: list[float]
        Weights for each channel for creating the synthetic input with summation.
    patch_extractor: PatchExtractor
        The patch extractor the patches were extracted from. Used for additional
        metadata.

    Returns
    -------
    input_region: UncorrelatedRegionData
        The input patch and its metadata, the data has the dimension L(Z)YX.
    target_region: UncorrelatedRegionData
        The target patch and its metadata, the data has the dimensions C(Z)YX.
    """
    ndims = len(patches.shape) - 1
    alpha_broadcast = np.array(alphas)[:, *(np.newaxis for _ in range(ndims))]
    # weight channels by alphas then sum on the channel axis
    # input dims will be L(Z)YX
    input_patch = (alpha_broadcast * patches).sum(axis=0)
    target_patch = patches[:, 0, ...]  # first L patch

    input_stacks = [
        patch_extractor.image_stacks[patch_spec["data_idx"]]
        for patch_spec in patch_specs
    ]
    source = [str(stack.source) for stack in input_stacks]
    data_shape = [stack.data_shape for stack in input_stacks]
    dtype = [str(stack.data_dtype) for stack in input_stacks]

    input_region = UncorrelatedRegionData(
        data=input_patch,
        source=source,
        data_shape=data_shape,
        dtype=dtype,
        region_spec=patch_specs,
        axes=axes,
    )
    target_region = UncorrelatedRegionData(
        data=target_patch,
        source=source,
        data_shape=data_shape,
        dtype=dtype,
        region_spec=patch_specs,
        axes=axes,
    )
    return input_region, target_region


def get_random_channel_patches(
    idx: int,  # TODO: is this needed it makes it work the same as original dataset
    patch_extractor: PatchExtractor[ImageStack],
    patching_strategy: PatchingStrategy,
    rng: np.random.Generator | None,
) -> tuple[NDArray[Any], list[PatchSpecs]]:
    """
    Select patches form random patch locations for each channel.

    Parameters
    ----------
    idx: int
        The dataset index.
    patch_extractor: PatchExtractor
        Used to extract patches from the data.
    patching_strategy: PatchingStrategy
        Patch locations will be sampled using the patching strategy.
    rng: numpy.random.Generator | None
        Useful for seeding the process. If `None` the default random number generator
        will be used.

    Returns
    -------
    NDArray[Any]
        The resulting patches with dimensions LC(Z)YX, where L contains the lateral
        context at multiple scales.
    list[PatchSpecs]
        A list of patch specification, one for each channel.
    """
    if rng is None:
        rng = np.random.default_rng()

    n_channels = patch_extractor.n_channels

    # in the original dataset, new random indices are chosen for each channel
    # the other channels can come from anywhere in the entire dataset
    indices = (idx, *rng.integers(patching_strategy.n_patches, size=(n_channels - 1)))

    # get n different patch specs for n different channels
    patch_specs = [patching_strategy.get_patch_spec(i) for i in indices]
    patches = extract_microsplit_patch(patch_extractor, patch_specs)

    return patches, patch_specs


# TODO: better name
def get_empty_channel_patches(
    idx: int,
    patch_extractor: PatchExtractor,
    patching_strategy: PatchingStrategy,
    signal_channels: dict[int, PatchFilterProtocol],
    empty_channels: dict[int, PatchFilterProtocol],
    patience: int,
    rng: np.random.Generator | None,
) -> tuple[NDArray[Any], list[PatchSpecs]]:
    """
    Select patches, specifying which channels should have signal and which should not.

    Parameters
    ----------
    idx: int
        The dataset index.
    patch_extractor: PatchExtractor
        Used to extract patches from the data.
    patching_strategy: PatchingStrategy
        Patch locations will be sampled using the patching strategy.
    signal_channels: dict[int, PatchFilterProtocol]
        A dictionary to specify the channels that should have signal and how they should
        be filtered. The keys are the channel index and the values are the patch filters
        used to determine if the channel patch is empty or not.
    empty_channels: dict[int, PatchFilterProtocol]
        A dictionary to specify the channels that should not have signal. Similar to
        the `signal_channels`.
    patience: int
        New patches are selected at random until a patch with signal or without is
        found, the `patience` determines how many times to look before giving up.
    rng: numpy.random.Generator | None
        Useful for seeding the process. If `None` the default random number generator
        will be used.

    Returns
    -------
    NDArray[Any]
        The resulting patches with dimensions LC(Z)YX, where L contains the lateral
        context at multiple scales.
    list[PatchSpecs]
        A list of patch specification, one for each channel.
    """
    if rng is None:
        rng = np.random.default_rng()

    # if a channel is not selected to be empty or filled it will from idx
    filled = set(signal_channels.keys())
    empty = set(empty_channels.keys())
    if len(intersect := filled.intersection(empty)) != 0:
        raise ValueError(
            "Channels cannot be selected as both empty and filled, the following "
            f"channels were selected as both {intersect}."
        )

    n_channels = patch_extractor.n_channels

    # start with random initial patches
    patches, patch_specs = get_random_channel_patches(
        idx, patch_extractor, patching_strategy, rng
    )

    # for each channel sample patches until they are empty or not empty
    for c in range(n_channels):

        # criterion for the while loop
        criterion: Callable[[NDArray[Any]], bool]
        filter_: PatchFilterProtocol
        if c in empty_channels:
            filter_ = empty_channels[c]
            criterion = is_not_empty(filter_)
        elif c in signal_channels:
            filter_ = signal_channels[c]
            criterion = is_empty(filter_)
        else:
            break

        patch = patches[c]
        patch_spec = patch_specs[c]
        patience_ = patience
        # only check if primary input is empty
        while criterion(patch[0]) and patience_ > 0:
            # sample random indices from anywhere in the dataset
            new_idx = rng.integers(patching_strategy.n_patches)
            patch_spec = patching_strategy.get_patch_spec(new_idx.item())
            patch = patch_extractor.extract_channel_patch(
                data_idx=patch_spec["data_idx"],
                sample_idx=patch_spec["sample_idx"],
                channels=[c],
                coords=patch_spec["coords"],
                patch_size=patch_spec["patch_size"],
            )[0]
            # ^ removing channel dim
            patience_ -= 1
        if patience <= 0:
            # TODO: log properly
            print(f"Out of patience finding patch for channel {c}")

        patches[c] = patch
        patch_specs[c] = patch_spec

    return patches, patch_specs


def extract_microsplit_patch(
    patch_extractor: PatchExtractor[ImageStack],
    patch_specs: PatchSpecs | list[PatchSpecs],
) -> NDArray[Any]:
    """
    Extract a MicroSplit patch with the dimensions LC(Z)YX.

    This patch can be used to synthesis an input patch by summing the C dimension, and
    it can be used to create a target patch by selecting the primary input from the
    L dimension, where L is to store lateral context patches.

    Parameters
    ----------
    patch_extractor: PatchExtractor
        Used to extract patches from the data.
    patch_specs: PatchSpec | list[PatchSpecs]
        A patch specification or a list of patch specifications — one for each channel.
        Different patch specs can be used or each channel to create uncorrelated channel
        patches.

    Returns
    -------
    NDArray[Any]
        The resulting patches with dimensions LC(Z)YX, where L contains the lateral
        context at multiple scales.
    """
    if isinstance(patch_specs, list):
        patches = np.concatenate(
            [
                patch_extractor.extract_channel_patch(
                    data_idx=patch_spec["data_idx"],
                    sample_idx=patch_spec["sample_idx"],
                    channels=[c],
                    coords=patch_spec["coords"],
                    patch_size=patch_spec["patch_size"],
                )
                for c, patch_spec in enumerate(patch_specs)
            ],
            axis=0,
        )
    else:
        patches = patch_extractor.extract_patch(
            data_idx=patch_specs["data_idx"],
            sample_idx=patch_specs["sample_idx"],
            coords=patch_specs["coords"],
            patch_size=patch_specs["patch_size"],
        )
    # Add L dimension if not present
    n_spatial_dims = patch_extractor.n_spatial_dims
    lateral_context_present = len(patches.shape) - n_spatial_dims == 2
    if not lateral_context_present:
        # insert a L dim
        patches = patches[:, np.newaxis]

    return patches



from collections.abc import Sequence
from typing import Any, Literal, Protocol

import numpy as np
from numpy.typing import NDArray
from skimage.transform import resize

from ..image_stack import ImageStack


class PatchConstructor(Protocol):
    """
    A callable that modifies how patches are constructed in the PatchExtractor.

    This protocol defines the signature of a callable that is passed as an argument to
    the `PatchExtractor`. It can be used to modify how patches are constructed, for
    example creating patches with multiple lateral context levels for MicroSplit.
    """

    def __call__(
        self,
        image_stack: ImageStack,
        sample_idx: int,
        channels: Sequence[int] | None,  # `channels = None` to select all channels
        coords: Sequence[int],
        patch_size: Sequence[int],
    ) -> NDArray[Any]:
        """
        Parameters
        ----------
        image_stack: ImageStack
            The image stack to construct a patch from.
        sample_idx: int
            Sample index. The first dimension of the image data will be indexed at this
            value.
        coords: Sequence of int
            The coordinates that define the start of a patch.
        patch_size: Sequence of int
            The size of the patch in each spatial dimension.

        Returns
        -------
        numpy.ndarray
            The patch.
        """
        ...


def default_patch_constr(
    image_stack: ImageStack,
    sample_idx: int,
    channels: Sequence[int] | None,  # `channels = None` to select all channels
    coords: Sequence[int],
    patch_size: Sequence[int],
) -> NDArray[Any]:
    return image_stack.extract_patch(
        sample_idx=sample_idx,
        channels=channels,
        coords=coords,
        patch_size=patch_size,
    )


# closure to create constructor funcs with particular multiscale_count and padding mode
def lateral_context_patch_constr(
    # TODO: will we stick with this as the parameter name
    multiscale_count: int,
    # TODO: add other modes?
    padding_mode: Literal["reflect", "wrap"],
) -> PatchConstructor:
    """
    Create a lateral context `PatchConstructor` for MicroSplit.

    Parameters
    ----------
    multiscale_count : int
        The number of multiscale inputs that will be created including the original
        image size.
    padding_mode : {"reflect", "wrap"}
        How lateral context inputs will be padded at the edge of the image. See
        [`numpy.pad`](https://numpy.org/devdocs/reference/generated/numpy.pad.html) for
        more information.

    Returns
    -------
    PatchConstructor
        The patch constructor function. It will return patches with the dimensions
        (C, L, (Z), Y, X) where L will be equal to `multiscale_count`, C is the number
        of channels in the image, and (Z), Y, X are the patch size.
    """

    def constructor_func(
        image_stack: ImageStack,
        sample_idx: int,
        channels: Sequence[int] | None,  # `channels = None` to select all channels
        coords: Sequence[int],
        patch_size: Sequence[int],
    ) -> NDArray[Any]:
        if channels is not None and len(channels) > 1:
            raise NotImplementedError(
                "Selecting multiple channels is currently not implemented for lateral "
                "context patches. Select a single channel or pass `channels=None` to "
                "select all channels."
            )

        shape = image_stack.data_shape
        spatial_shape = shape[2:]
        n_channels = shape[1] if channels is None else 1

        # There will now be an additional lc dimension,
        # this has to be handled correctly by the dataset
        # TODO: maybe we want to limit this constructor to only images with 1 channel
        #   then we can put LCs in the channel dimension
        #   but not sure if this artificially limits potential use-cases
        patch = np.zeros((n_channels, multiscale_count, *patch_size))
        for scale in range(multiscale_count):
            lc_patch_size = np.array(patch_size) * (2**scale)
            lc_start = np.array(coords) + np.array(patch_size) // 2 - lc_patch_size // 2
            lc_end = lc_start + np.array(lc_patch_size)

            start_clipped = np.clip(
                lc_start, np.zeros_like(spatial_shape), np.array(spatial_shape)
            )
            end_clipped = np.clip(
                lc_end, np.zeros_like(spatial_shape), np.array(spatial_shape)
            )
            size_clipped = end_clipped - start_clipped

            lc_patch = image_stack.extract_patch(
                sample_idx, channels, start_clipped, size_clipped
            )
            pad_before = start_clipped - lc_start
            pad_after = lc_end - end_clipped
            pad_width = np.concatenate(
                [
                    # zeros to not pad the channel axis
                    np.zeros((1, 2), dtype=int),
                    np.stack([pad_before, pad_after], axis=-1),
                ]
            )
            lc_patch = np.pad(
                lc_patch,
                pad_width,
                mode=padding_mode,
            )
            # TODO: test different downscaling? skimage suggests downscale_local_mean
            lc_patch = resize(lc_patch, (n_channels, *patch_size))
            patch[:, scale, ...] = lc_patch
        return patch

    return constructor_func



"""A module to contain type definitions relating to patching strategies."""

from collections.abc import Sequence
from typing import Protocol, TypedDict, TypeGuard, TypeVar

RegionSpecs = TypeVar("RegionSpecs", bound="PatchSpecs")


class PatchSpecs(TypedDict):
    """A dictionary that specifies a single patch in a series of `ImageStacks`.

    Attributes
    ----------
    data_idx: int
        Determines which `ImageStack` a patch belongs to, within a series of
        `ImageStack`s.
    sample_idx: int
        Determines which sample a patch belongs to, within an `ImageStack`.
    coords: sequence of int
        The top-left (and first z-slice for 3D data) of a patch. The sequence will have
        length 2 or 3, for 2D and 3D data respectively.
    patch_size: sequence of int
        The size of the patch. The sequence will have length 2 or 3, for 2D and 3D data
        respectively.
    """

    data_idx: int
    sample_idx: int
    coords: Sequence[int]
    patch_size: Sequence[int]


class TileSpecs(PatchSpecs):
    """A dictionary that specifies a single patch in a series of `ImageStacks`.

    Attributes
    ----------
    data_idx: int
        Determines which `ImageStack` a patch belongs to, within a series of
        `ImageStack`s.
    sample_idx: int
        Determines which sample a patch belongs to, within an `ImageStack`.
    coords: sequence of int
        The top-left (and first z-slice for 3D data) of a patch. The sequence will have
        length 2 or 3, for 2D and 3D data respectively.
    patch_size: sequence of int
        The size of the patch. The sequence will have length 2 or 3, for 2D and 3D data
        respectively.
    crop_coords: sequence of int
        The top-left side of where the tile will be cropped, in coordinates relative
        to the tile.
    crop_size: sequence of int
        The size of the cropped tile.
    stitch_coords: sequence of int
        Where the tile will be stitched back into an image, taking into account
        that the tile will be cropped, in coords relative to the image.
    total_tiles: int
        Number of tiles belonging to the same data.
    """

    crop_coords: Sequence[int]
    crop_size: Sequence[int]
    stitch_coords: Sequence[int]
    total_tiles: int


def is_tile_specs(specs: PatchSpecs) -> TypeGuard[TileSpecs]:
    """Determine whether a given PatchSpecs is a TileSpecs.

    Used for type checking.

    Parameters
    ----------
    specs : PatchSpecs
        A patch specification.

    Returns
    -------
    bool
        Whether the given specs is a TileSpecs.
    """
    return (
        ("crop_coords" in specs)
        and ("crop_size" in specs)
        and ("stitch_coords" in specs)
    )


class PatchingStrategy(Protocol):
    """
    An interface for patching strategies.

    Patching strategies are a component of the `CAREamicsDataset`; they determine
    how patches are extracted from the underlying data.

    Attributes
    ----------
    n_patches: int
        The number of patches that the patching strategy will return.

    Methods
    -------
    get_patch_spec(index: int) -> PatchSpecs
        Get a patch specification for a given patch index.
    """

    @property
    def n_patches(self) -> int:
        """
        The number of patches that the patching strategy will return.

        It also determines the maximum index that can be given to `get_patch_spec`,
        and the length of the `CAREamicsDataset`.

        Returns
        -------
        int
            Number of patches.
        """
        ...

    def get_patch_spec(self, index: int) -> PatchSpecs:
        """
        Get a patch specification for a given patch index.

        This method is intended to be called from within the
        `CAREamicsDataset.__getitem__`. The index will be passed through from this
        method.

        Parameters
        ----------
        index : int
            A patch index.

        Returns
        -------
        PatchSpecs
            A dictionary that specifies a single patch in a series of `ImageStacks`.
        """
        ...

    # Note: this is used by the FileIterSampler
    def get_patch_indices(self, data_idx: int) -> Sequence[int]:
        """
        Get the patch indices will return patches for a specific `image_stack`.

        The `image_stack` corresponds to the given `data_idx`.

        Parameters
        ----------
        data_idx : int
            An index that corresponds to a given `image_stack`.

        Returns
        -------
        sequence of int
            A sequence of patch indices, that when used to index the `CAREamicsDataset
            will return a patch that comes from the `image_stack` corresponding to the
            given `data_idx`.
        """
        ...



import numpy as np
import pytest
from numpy.typing import NDArray

from careamics.dataset_ng.image_stack import InMemoryImageStack
from careamics.dataset_ng.image_stack.image_utils import channel_slice


@pytest.fixture
def array_stack(shape, axes) -> tuple[NDArray, InMemoryImageStack]:
    data = np.arange(np.prod(shape)).reshape(shape)
    return data, InMemoryImageStack.from_array(data, axes)


# TODO test reshaping of array


@pytest.mark.parametrize(
    "shape, axes",
    [
        ((2, 3, 64, 64), "SCYX"),
        ((3, 4, 32, 32, 32), "SCZYX"),
    ],
)
def test_from_array(array_stack):
    data, image_stack = array_stack
    np.testing.assert_array_equal(image_stack._data, data)
    assert image_stack.data_shape == data.shape
    assert image_stack.data_dtype == data.dtype


@pytest.mark.parametrize(
    "shape, axes, channels",
    [
        ((2, 1, 64, 64), "SCYX", None),
        ((2, 1, 64, 64), "SCYX", [0]),
        ((2, 3, 64, 64), "SCYX", None),
        ((2, 3, 64, 64), "SCYX", [0, 2]),
        ((2, 3, 64, 64), "SCYX", [1]),
    ],
)
def test_extract_channels(array_stack, channels):
    data, image_stack = array_stack
    data_shape = data.shape

    patch = image_stack.extract_patch(
        sample_idx=1,
        channels=channels,
        coords=(10, 10),
        patch_size=(32, 32),
    )
    assert len(patch.shape) == 3  # no Z
    assert patch.shape[0] == len(channels) if channels is not None else data_shape[1]

    expected_patch = data[
        1,
        channel_slice(channels),
        10 : 10 + 32,
        10 : 10 + 32,
    ]
    np.testing.assert_array_equal(patch, expected_patch)


@pytest.mark.parametrize(
    "shape, axes, channels",
    [
        ((2, 3, 64, 64), "SCYX", [0, 4]),
        ((2, 3, 64, 64), "SCYX", [3]),
    ],
)
def test_extract_channel_error(array_stack, channels):
    data, image_stack = array_stack
    data_shape = data.shape

    expected_msg = (
        f"Channel index {channels[-1]} is out of bounds for data with "
        f"{data_shape[1]} channels. Check the provided `channels` "
        f"parameter in the configuration for erroneous channel "
        f"indices."
    )

    with pytest.raises(ValueError, match=expected_msg):
        image_stack.extract_patch(
            sample_idx=0,
            channels=channels,
            coords=(0, 0),
            patch_size=(16, 16),
        )



import numpy as np
import pytest
import tifffile

from careamics.config import InferenceConfig
from careamics.dataset import IterableTiledPredDataset


# TODO extract tiles is returning C(Z)YX and no singleton S!
@pytest.mark.parametrize(
    "n_files, shape, axes, expected_shape",
    [
        (1, (16, 16), "YX", (1, 16, 16)),
        (1, (3, 16, 16), "CYX", (3, 16, 16)),
        (1, (8, 16, 16), "ZYX", (1, 8, 16, 16)),
        (1, (3, 8, 16, 16), "CZYX", (3, 8, 16, 16)),
        (1, (4, 16, 16), "SYX", (1, 16, 16)),
        (1, (4, 3, 16, 16), "SCYX", (3, 16, 16)),
        (1, (4, 3, 8, 16, 16), "SCZYX", (3, 8, 16, 16)),
        (3, (16, 16), "YX", (1, 16, 16)),
        (3, (3, 16, 16), "CYX", (3, 16, 16)),
        (3, (8, 16, 16), "ZYX", (1, 8, 16, 16)),
        (3, (3, 8, 16, 16), "CZYX", (3, 8, 16, 16)),
        (3, (4, 16, 16), "SYX", (1, 16, 16)),
        (3, (4, 3, 16, 16), "SCYX", (3, 16, 16)),
        (3, (4, 3, 8, 16, 16), "SCZYX", (3, 8, 16, 16)),
    ],
)
def test_correct_normalized_outputs(tmp_path, n_files, shape, axes, expected_shape):
    """Test that the dataset returns normalized images with singleton
    sample dimension."""
    rng = np.random.default_rng(42)

    tile_size = (8, 8, 8) if "Z" in axes else (8, 8)
    tile_overlap = (4, 4, 4) if "Z" in axes else (4, 4)

    # check expected length
    n_tiles = np.prod(
        np.ceil(
            (expected_shape[1:] - np.array(tile_overlap))
            / (np.array(tile_size) - np.array(tile_overlap))
        )
    ).astype(int)

    # check number of samples
    if "S" in axes:
        # get index
        idx = axes.index("S")
        n_samples = shape[idx]
    else:
        n_samples = 1

    # check number of channels
    if "C" in axes:
        # get index
        idx = axes.index("C")
        n_channels = shape[idx]
    else:
        n_channels = 1

    # create array
    new_shape = (n_files,) + shape
    array = 255 * rng.random(new_shape)

    # create config
    config = InferenceConfig(
        data_type="tiff",
        axes=axes,
        image_means=[np.mean(array)] * n_channels,
        image_stds=[np.std(array)] * n_channels,
        tile_size=tile_size,
        tile_overlap=tile_overlap,
    )

    files = []
    for i in range(n_files):
        file = tmp_path / f"file_{i}.tif"
        tifffile.imwrite(file, array[i])
        files.append(file)

    # create dataset
    dataset = IterableTiledPredDataset(config, files)

    # get all images
    dataset = list(dataset)

    # check length
    assert len(dataset) == n_files * n_samples * n_tiles

    # check that the dataset returns normalized images
    for i in range(len(dataset)):
        img, _ = dataset[i]
        img = img[0]
        # check that it has the correct shape
        assert img.shape == (n_channels,) + tile_size

        # check that the image is normalized
        assert np.isclose(np.mean(img), 0, atol=0.25)
        assert np.isclose(np.std(img), 1, atol=0.2)

        # check that they are independent slices
        for j in range(i + 1, len(dataset)):
            img2, _ = dataset[j]
            assert not np.allclose(img, img2)



"""Tiled patching Pydantic model."""

from collections.abc import Sequence
from typing import Literal

from pydantic import Field

from ._overlapping_patched_config import _OverlappingPatchedConfig


# TODO with UNet tiling must obey different rules than sequential tiling
#   - needs to validated at the level of the configuration
class TiledPatchingConfig(_OverlappingPatchedConfig):
    """Tiled patching Pydantic model.

    Attributes
    ----------
    name : "tiled"
        The name of the patching strategy.
    patch_size : sequence of int
        The size of the patch in each spatial dimension, each patch size must be a power
        of 2 and larger than 8.
    overlaps : sequence of int
        The overlaps between patches in each spatial dimension. The overlaps must be
        smaller than the patch size in each spatial dimension, and the number of
        dimensions be either 2 or 3.
    """

    name: Literal["tiled"] = "tiled"
    """The name of the patching strategy."""

    overlaps: Sequence[int] = Field(
        ...,
        min_length=2,
        max_length=3,
    )
    """The overlaps between patches in each spatial dimension. The overlaps must be
    smaller than the patch size in each spatial dimension, and the number of dimensions
    be either 2 or 3.
    """



import numpy as np

from careamics.dataset_ng.image_stack_loader import load_arrays
from careamics.dataset_ng.patch_extractor import PatchExtractor
from careamics.dataset_ng.patch_filter import MaskCoordFilter
from careamics.dataset_ng.patching_strategies import PatchSpecs

# TODO test probability of application


def test_filter():
    """Test MaskPatchFilter functionality."""
    size = 16
    mask = np.zeros((size, size))
    mask[size // 4 : -size // 4, size // 4 : -size // 4] = 1

    image_stacks = load_arrays(source=[mask], axes="YX")
    mask_filter = MaskCoordFilter(
        mask_extractor=PatchExtractor(image_stacks),
        coverage=0.50,
    )
    assert mask_filter.filter_out(
        PatchSpecs(data_idx=0, sample_idx=0, coords=[0, 0], patch_size=[4, 4])
    )  # corner, no overlap with mask
    assert mask_filter.filter_out(
        PatchSpecs(data_idx=0, sample_idx=0, coords=[2, 2], patch_size=[4, 4])
    )  # 25% overlap
    assert not mask_filter.filter_out(
        PatchSpecs(data_idx=0, sample_idx=0, coords=[2, 4], patch_size=[4, 4])
    )  # 50% overlap
    assert not mask_filter.filter_out(
        PatchSpecs(data_idx=0, sample_idx=0, coords=[4, 4], patch_size=[4, 4])
    )  # 100% overlap



import random

import numpy as np
import pytest
from torch.utils.data.dataloader import default_collate

from careamics.dataset_ng.dataset import ImageRegionData
from careamics.lightning.dataset_ng.prediction import (
    combine_samples,
    convert_prediction,
    decollate_image_region_data,
)


@pytest.fixture
def batches(source_name: str) -> list[ImageRegionData]:
    """
    Fixture providing batches of `ImageRegionData` to test decollate.

    Returns
    -------
    list of ImageRegionData
        List of batches of `ImageRegionData`.
    """
    batches = []
    for b in range(5):  # data idx spread over multiple batches
        batch = []
        for i in range(4):
            data_idx = (b * 4 + i) // 10

            batch.append(
                ImageRegionData(
                    source=f"{data_idx}.tiff" if source_name == "file" else "array",
                    data=(
                        data_idx * np.ones((1, 32, 32)).astype(np.float32)
                    ),  # B dim added by collate
                    data_shape=(10, 32, 32),
                    dtype="float32",
                    axes="SYX",
                    # non-sense to check that they are properly decollated
                    region_spec={
                        "data_idx": data_idx,
                        "sample_idx": (b * 4 + i) % 10,
                        "coords": (0, i * 4),
                        "patch_size": (4, i),
                    },
                    additional_metadata={
                        "chunks": (1, 1, 16, 16),
                        "shards": (1, 1, 32, 32),
                    },
                )
            )

        batch_collated = default_collate(batch)
        batches.append(batch_collated)
    return batches


class TestCombineSamples:

    @pytest.mark.parametrize("source_name", ["file"])  # injected in fixture
    def test_combine_prediction_by_data_idx(
        self, batches: list[ImageRegionData]
    ) -> None:
        """Test `combine_prediction_by_data_idx` function."""
        all_decollated: list[ImageRegionData] = []
        for batch in batches:
            decollated = decollate_image_region_data(batch)
            all_decollated.extend(decollated)

        combined_predictions, _ = combine_samples(all_decollated)
        assert len(combined_predictions) == 2
        assert combined_predictions[0].shape == (10, 32, 32)
        assert np.all(combined_predictions[0] == 0)
        assert combined_predictions[1].shape == (10, 32, 32)
        assert np.all(combined_predictions[1] == 1)

    @pytest.mark.parametrize("source_name", ["file"])  # injected in fixture
    def test_data_idx_order(self, batches: list[ImageRegionData]) -> None:
        """Test that `combine_prediction_by_data_idx` returns data in the correct
        order."""
        all_decollated: list[ImageRegionData] = []
        for batch in batches:
            decollated = decollate_image_region_data(batch)
            all_decollated.extend(decollated)

        data_indices = np.unique([b.region_spec["data_idx"] for b in all_decollated])
        assert len(data_indices) == 2
        assert data_indices[0] < data_indices[1]

        # shuffle decollated to test ordering
        random.shuffle(all_decollated)

        # predict and check ordering
        combined_predictions, sources = combine_samples(all_decollated)
        assert len(combined_predictions) == len(sources) == 2
        assert sources[0] == "0.tiff"
        assert sources[1] == "1.tiff"

        assert np.all(combined_predictions[0] == 0)  # data_idx = 0
        assert np.all(combined_predictions[1] == 1)  # data_idx = 1


@pytest.mark.parametrize("n_batch", [1, 2, 4])
def test_decollate_image_region_data(n_batch) -> None:
    """
    Test `decollate_image_region_data` function.

    Parameters
    ----------
    ordered_array : NDArray
        Ordered array fixture.
    """
    batch: list[ImageRegionData] = []
    for i in range(n_batch):
        batch.append(
            ImageRegionData(
                source="array.tiff",
                data=np.ones((1, 4, 4)).astype(np.float32),  # B dim is added by collate
                data_shape=(32, 32),
                dtype=str(np.float32),
                axes="YX",
                region_spec={
                    "data_idx": i,
                    "sample_idx": 0,
                    "coords": (0, i * 4),
                    "patch_size": (4, 4),
                },
                additional_metadata={
                    "chunks": (1, 1, 16, i),
                    "shards": (1, 1, 32, i * 2),
                },
            )
        )

    batch_collated = default_collate(batch)
    decollated = decollate_image_region_data(batch_collated)

    assert len(decollated) == len(batch)
    for i in range(len(batch)):
        np.testing.assert_array_equal(decollated[i].data, batch[i].data, verbose=True)
        assert decollated[i].source == batch[i].source
        assert decollated[i].data_shape == batch[i].data_shape
        assert decollated[i].dtype == batch[i].dtype
        assert decollated[i].axes == batch[i].axes
        assert decollated[i].region_spec == batch[i].region_spec
        assert (
            decollated[i].additional_metadata["chunks"]
            == batch[i].additional_metadata["chunks"]
        )
        assert (
            decollated[i].additional_metadata["shards"]
            == batch[i].additional_metadata["shards"]
        )


class TestConvertPrediction:

    @pytest.mark.parametrize("source_name", ["file"])  # injected in fixture
    def test_convert_arrays(self, batches: list[ImageRegionData]) -> None:
        """Test `convert_arrays` function."""
        predictions, sources = convert_prediction(batches, tiled=False)
        assert len(predictions) == 2  # 2 data idx in fixture
        assert predictions[0].shape == (10, 32, 32)
        assert predictions[1].shape == (10, 32, 32)
        assert sources == ["0.tiff", "1.tiff"]

    @pytest.mark.parametrize("source_name", ["array"])  # injected in fixture
    def test_convert_arrays_empty_list(self, batches: list[ImageRegionData]) -> None:
        """Test `convert_arrays` with "array" sources returns an empty list."""
        _, sources = convert_prediction(batches, tiled=False)
        assert sources == []



"""Module containing `PredictionWriterCallback` class."""

from __future__ import annotations

from collections.abc import Sequence
from pathlib import Path
from typing import Any

from pytorch_lightning import LightningModule, Trainer
from pytorch_lightning.callbacks import BasePredictionWriter

from careamics.dataset_ng.dataset import ImageRegionData
from careamics.file_io.write.get_func import SupportedWriteType, WriteFunc
from careamics.lightning.dataset_ng.prediction import decollate_image_region_data
from careamics.utils import get_logger

from .write_strategy import WriteStrategy
from .write_strategy_factory import create_write_strategy

logger = get_logger(__name__)


class PredictionWriterCallback(BasePredictionWriter):
    """
    PyTorch Lightning callback to save predictions.

    A `WriteStrategy` must be provided at instantiation or later via
    `set_writing_strategy`. This allows passing the callback to the Lightning `Trainer`
    before knowing what writing strategy (e.g. tiling or file tyope) will be used.

    By default the prediction writer is enabled, but it can be disabled by passing
    `enable_writing=False` at instantiation or by calling `enable_writing(False)`.
    Subsequently, writing can be re-enabled by calling `enable_writing(True)`.

    Parameters
    ----------
    dirpath : Path or str, default="predictions"
        The path to the directory where prediction outputs will be saved. If
        `dirpath` is not absolute it is assumed to be relative to current working
        directory.
    write_strategy : WriteStrategy or None, default=None
        A strategy for writing predictions.
    enable_writing : bool, default=True
        If writing predictions should be enabled by default.

    Attributes
    ----------
    is_enabled : bool
        Whether writing predictions is enabled.
    dirpath : pathlib.Path, default=""
        The path to the directory where prediction outputs will be saved. If
        `dirpath` is not absolute it is assumed to be relative to current working
        directory.
    write_strategy : WriteStrategy or None
            A strategy for writing predictions.
    """

    def __init__(
        self,
        dirpath: Path | str = "",
        write_strategy: WriteStrategy | None = None,
        enable_writing: bool = True,
    ):
        """
        Constructor.

        A `WriteStrategy` must be provided at instantiation or later via
        `set_writing_strategy`.

        Parameters
        ----------
        dirpath : pathlib.Path or str, default="predictions"
            The path to the directory where prediction outputs will be saved. If
            `dirpath` is not absolute it is assumed to be relative to current working
            directory.
        write_strategy : WriteStrategy or None, default=None
            A strategy for writing predictions.
        enable_writing : bool, default=True
            If writing predictions should be enabled by default.
        """
        super().__init__(write_interval="batch")

        self.is_enabled = enable_writing
        self.write_strategy = write_strategy

        self.dirpath: Path

        # if a dirpath is provided, initialize it
        # in some cases (e.g. zarr), destination is provided by the zarr store path
        if dirpath != "":
            self._init_dirpath(dirpath)

    def enable_writing(self, enable_writing: bool) -> None:
        """Enable or disable writing.

        Parameters
        ----------
        enable_writing : bool
            If writing predictions should be enabled.
        """
        self.is_enabled = enable_writing

    def _init_dirpath(self, dirpath):
        """
        Initialize directory path. Should only be called from `__init__`.

        Parameters
        ----------
        dirpath : pathlib.Path
            See `__init__` description.
        """
        dirpath = Path(dirpath)
        if not dirpath.is_absolute():
            dirpath = Path.cwd() / dirpath
            logger.warning(
                "Prediction output directory is not absolute, absolute path assumed to"
                f"be '{dirpath}'"
            )
        self.dirpath = dirpath

    def setup(self, trainer: Trainer, pl_module: LightningModule, stage: str) -> None:
        """
        Create the prediction output directory when predict begins.

        Called when fit, validate, test, predict, or tune begins.

        Parameters
        ----------
        trainer : Trainer
            PyTorch Lightning trainer.
        pl_module : LightningModule
            PyTorch Lightning module.
        stage : str
            Stage of training e.g. 'predict', 'fit', 'validate'.
        """
        super().setup(trainer, pl_module, stage)
        if stage == "predict":
            if self.dirpath is not None:
                # make prediction output directory
                logger.info("Making prediction output directory.")
                self.dirpath.mkdir(parents=True, exist_ok=True)

    def set_writing_strategy(
        self,
        write_type: SupportedWriteType,
        tiled: bool,
        write_func: WriteFunc | None = None,
        write_extension: str | None = None,
        write_func_kwargs: dict[str, Any] | None = None,
    ) -> None:
        """
        Set the writing strategy.

        Must be called before writing predictions.

        Parameters
        ----------
        write_type : SupportedWriteType
            The type of writing to perform.
        tiled : bool
            Whether to write in tiled format.
        write_func : WriteFunc or None, default=None
            A custom writing function.
        write_extension : str or None, default=None
            The file extension to use when writing files.
        write_func_kwargs : dict of str to Any, default=None
            Additional keyword arguments to pass to `write_func`.
        """
        self.write_strategy = create_write_strategy(
            write_type=write_type,
            tiled=tiled,
            write_func=write_func,
            write_extension=write_extension,
            write_func_kwargs=write_func_kwargs,
        )

    def write_on_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        prediction: ImageRegionData,
        batch_indices: Sequence[int] | None,
        batch: ImageRegionData,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        """
        Write predictions at the end of a batch.

        Writing method is determined by the attribute `write_strategy`.

        Parameters
        ----------
        trainer : Trainer
            PyTorch Lightning trainer.
        pl_module : LightningModule
            PyTorch Lightning module.
        prediction : ImageRegionData
            Prediction outputs of `batch`.
        batch_indices : sequence of Any, optional
            Batch indices.
        batch : ImageRegionData
            Input batch.
        batch_idx : int
            Batch index.
        dataloader_idx : int
            Dataloader index.
        """
        # if writing prediction is turned off
        if not self.is_enabled:
            return

        if self.write_strategy is not None:
            assert prediction is not None
            predictions = decollate_image_region_data(prediction)

            self.write_strategy.write_batch(
                dirpath=self.dirpath,
                predictions=predictions,
            )
        else:
            raise RuntimeError(
                "No write strategy defined for `PredictionWriterCallback`, cannot write"
                " predictions. Call `set_writing_strategy` to pass a write strategy."
            )



#! /usr/bin/env python
"""Simple test used to check CAREamist import after packaging in the deployment CI."""


def test_smoke():
    """Import careamics.CAREamist."""
    from careamics import CAREamist
    from careamics.config import create_n2v_configuration

    # create configuration
    config = create_n2v_configuration(
        experiment_name="smoke test",
        data_type="array",
        axes="YX",
        patch_size=(64, 64),
        batch_size=8,
    )
    CAREamist(source=config)


if __name__ == "__main__":
    test_smoke()



"""Filter using an image mask."""

import numpy as np

from careamics.dataset_ng.image_stack import GenericImageStack
from careamics.dataset_ng.patch_extractor import PatchExtractor
from careamics.dataset_ng.patch_filter.coordinate_filter_protocol import (
    CoordinateFilterProtocol,
)
from careamics.dataset_ng.patching_strategies import PatchSpecs


# TODO is it more intuitive to have a negative mask? (mask of what to avoid)
class MaskCoordFilter(CoordinateFilterProtocol):
    """
    Filter patch coordinates based on an image mask.

    Attributes
    ----------
    mask_extractor : PatchExtractor[GenericImageStack]
        Patch extractor for the binary mask to use for filtering.
    coverage_perc : float
        Minimum percentage of masked pixels required to keep a patch.
    p : float
        Probability of applying the filter to a patch.
    rng : np.random.Generator
        Random number generator for stochastic filtering.
    """

    def __init__(
        self,
        mask_extractor: PatchExtractor[GenericImageStack],
        coverage: float,
        p: float = 1.0,
        seed: int | None = None,
    ) -> None:
        """
        Create a MaskCoordFilter.

        This filter removes patches who fall below a threshold of masked pixels
        percentage. The mask is expected to be a positive mask where masked pixels
        correspond to regions of interest.

        Parameters
        ----------
        mask_extractor : PatchExtractor[GenericImageStack]
            The patch extractor for the mask used for filtering.
        coverage : float
            Minimum percentage of masked pixels required to keep a patch. Must be
            between 0 and 1.
        p : float, default=1
            Probability of applying the filter to a patch. Must be between 0 and 1.
        seed : int | None, default=None
            Seed for the random number generator for reproducibility.

        Raises
        ------
        ValueError
            If coverage is not between 0 and 1.
        ValueError
            If p is not between 0 and 1.
        """

        if not (0 <= coverage <= 1):
            raise ValueError("Probability p must be between 0 and 1.")
        if not (0 <= p <= 1):
            raise ValueError("Probability p must be between 0 and 1.")

        self.mask_extractor = mask_extractor
        self.coverage = coverage

        self.p = p
        self.rng = np.random.default_rng(seed)

    def filter_out(self, patch_specs: PatchSpecs) -> bool:
        """
        Determine whether to filter out a patch based an image mask.

        Parameters
        ----------
        patch : PatchSpecs
            The patch coordinates to evaluate.

        Returns
        -------
        bool
            True if the patch should be filtered out, False otherwise.
        """

        if self.rng.uniform(0, 1) < self.p:
            mask_patch = self.mask_extractor.extract_patch(**patch_specs)

            masked_fraction = np.sum(mask_patch) / mask_patch.size
            if masked_fraction < self.coverage:
                return True
        return False



"""Module for the `GroupedIndexSampler`."""

from collections.abc import Iterator, Sequence
from typing import Self

import numpy as np
from numpy.random import Generator, default_rng
from torch.utils.data import Sampler

from careamics.dataset_ng.dataset import CareamicsDataset


class GroupedIndexSampler(Sampler):
    """
    A PyTorch Sampler iterates through groups of indices.

    The order of the groups will be shuffled and the order of the indices within the
    groups will be shuffled.

    This sampler is useful for iterative file loading — one file should be loaded at a
    time so indices belonging to the same file should be grouped, but the order of the
    files and the order of the indices should be shuffled.
    """

    def __init__(self, grouped_indices: Sequence[Sequence[int]], rng: Generator | None):
        """
        Parameters
        ----------
        grouped_indices : Sequence of (Sequence of int)
            The indices that should be iterated through in groups.
        """
        super().__init__()
        if rng is None:
            self.rng = default_rng()
        else:
            self.rng = rng
        # TODO: validate indices are unique across groups
        self.grouped_indices = grouped_indices

    @classmethod
    def from_dataset(
        cls, dataset: CareamicsDataset, rng: Generator | None = None
    ) -> Self:
        """
        Create the sampler from a CareamicsDataset.

        The grouped indices will be retrieved from the dataset's patching strategy.

        Parameters
        ----------
        dataset: CareamicsDataset
            An instance of the CareamicsDataset to create the sampler for.
        rng: numpy.random.Generator, optional
            Numpy random number generator that can be used to seed the sampler.
        """
        n_data_samples = len(dataset.input_extractor.shapes)
        grouped_indices: list[Sequence[int]] = [
            dataset.patching_strategy.get_patch_indices(i)
            for i in range(n_data_samples)
        ]
        return cls(grouped_indices=grouped_indices, rng=rng)

    def __iter__(self) -> Iterator[int]:

        # shuffle the groups and the sub groups but keep indices in a group adjacent
        group_order = np.arange(len(self.grouped_indices))
        self.rng.shuffle(group_order)
        for group_idx in group_order:
            group = self.grouped_indices[group_idx.item()]
            index_order = np.arange(len(group))
            self.rng.shuffle(index_order)
            for idx in index_order:
                yield group[idx.item()]



"""Data configuration."""

from __future__ import annotations

import os
import random
import sys
from collections.abc import Sequence
from pprint import pformat
from typing import Annotated, Any, Literal, Self, Union
from warnings import warn

import numpy as np
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    PlainSerializer,
    ValidationInfo,
    field_validator,
    model_validator,
)

from careamics.utils import BaseEnum

from ..transformations import XYFlipConfig, XYRandomRotate90Config
from ..validators import check_axes_validity, check_czi_axes_validity
from .normalization_config import NormalizationConfig
from .patch_filter import (
    MaskFilterConfig,
    MaxFilterConfig,
    MeanSTDFilterConfig,
    ShannonFilterConfig,
)
from .patching_strategies import (
    FixedRandomPatchingConfig,
    RandomPatchingConfig,
    StratifiedPatchingConfig,
    TiledPatchingConfig,
    WholePatchingConfig,
)

# TODO: Validate the specific sizes of tiles and overlaps given UNet constraints
#   - needs to be done in the Configuration
#   - patches and overlaps sizes must also be checked against dimensionality
#   - Should we have a UNet and a LVAE NGDataConfig subclass with specific validations?

# TODO: is 3D updated anywhere in the code in CAREamist/downstream?
#       - this will be important when swapping the data config in Configuration
#       - `set_3D` currently not implemented here

# TODO: this module is very long, can we split the validation somewhere else and
#       leverage Pydantic to add validation directly to the declaration of each field?


def generate_random_seed() -> int:
    """Generate a random seed for reproducibility.

    Returns
    -------
    int
        A random integer between 1 and 2^31 - 1.
    """
    return random.randint(1, 2**31 - 1)


def np_float_to_scientific_str(x: float) -> str:
    """Return a string scientific representation of a float.

    In particular, this method is used to serialize floats to strings, allowing
    numpy.float32 to be passed in the Pydantic model and written to a yaml file as str.

    Parameters
    ----------
    x : float
        Input value.

    Returns
    -------
    str
        Scientific string representation of the input value.
    """
    return np.format_float_scientific(x, precision=7)


Float = Annotated[float, PlainSerializer(np_float_to_scientific_str, return_type=str)]
"""Annotated float type, used to serialize floats to strings."""

PatchingConfig = Union[
    FixedRandomPatchingConfig,
    RandomPatchingConfig,
    StratifiedPatchingConfig,
    TiledPatchingConfig,
    WholePatchingConfig,
]
"""Patching strategy type."""

PatchFilterConfig = Union[
    MaxFilterConfig,
    MeanSTDFilterConfig,
    ShannonFilterConfig,
]
"""Patch filter type."""

CoordFilterConfig = Union[MaskFilterConfig]  # add more here as needed
"""Coordinate filter type."""


class Mode(str, BaseEnum):
    """Dataset mode."""

    TRAINING = "training"
    VALIDATING = "validating"
    PREDICTING = "predicting"


def default_in_memory(validated_params: dict[str, Any]) -> bool:
    """Default factory for the `in_memory` field.

    Based on the value of `data_type`, set the default for `in_memory` to `True` if
    the data type is 'array', 'tiff', or 'custom', and to `False` otherwise (`zarr`
    or 'czi').

    Parameters
    ----------
    validated_params : dict of {str: Any}
        Validated parameters.

    Returns
    -------
    bool
        Default value for the `in_memory` field.
    """
    return validated_params.get("data_type") not in ("zarr", "czi")


class NGDataConfig(BaseModel):
    """Next-Generation Dataset configuration.

    NGDataConfig are used for both training and prediction, with the patching strategy
    determining how the data is processed. Note that `random` is the only patching
    strategy compatible with training, while `tiled` and `whole` are only used for
    prediction.

    All supported transforms are defined in the SupportedTransform enum.
    """

    # Pydantic class configuration
    model_config = ConfigDict(
        validate_assignment=True,
    )

    # Dataset configuration
    mode: Mode
    """Dataset mode, either training, validating or predicting."""

    data_type: Literal["array", "tiff", "zarr", "czi", "custom"]
    """Type of input data."""

    axes: str
    """Axes of the data, as defined in SupportedAxes."""

    # TODO: update docs for stratified patching
    patching: PatchingConfig = Field(..., discriminator="name")
    """Patching strategy to use. Note that `random` is the only supported strategy for
    training, while `tiled` and `whole` are only used for prediction."""

    normalization: NormalizationConfig = Field(...)
    """Normalization configuration to use."""

    # Optional fields
    batch_size: int = Field(default=1, ge=1, validate_default=True)
    """Batch size for training."""

    in_memory: bool = Field(default_factory=default_in_memory, validate_default=True)
    """Whether to load all data into memory. This is only supported for 'array',
    'tiff' and 'custom' data types. Must be `True` for `array`. If `None`, defaults to
    `True` for 'array', 'tiff' and `custom`, and `False` for 'zarr' and 'czi' data
    types."""

    channels: Sequence[int] | None = Field(default=None)
    """Channels to use from the data. If `None`, all channels are used."""

    patch_filter: PatchFilterConfig | None = Field(default=None, discriminator="name")
    """Patch filter to apply when using random patching. Only available if
    mode is `training`."""

    coord_filter: CoordFilterConfig | None = Field(default=None, discriminator="name")
    """Coordinate filter to apply when using random patching. Only available if
    mode is `training`."""

    patch_filter_patience: int = Field(default=5, ge=1)
    """Number of consecutive patches not passing the filter before accepting the next
    patch."""

    transforms: Sequence[Union[XYFlipConfig, XYRandomRotate90Config]] = Field(
        default=(
            XYFlipConfig(),
            XYRandomRotate90Config(),
        ),
        validate_default=True,
    )
    """List of transformations to apply to the data, available transforms are defined
    in SupportedTransform."""

    train_dataloader_params: dict[str, Any] = Field(
        default={"shuffle": True}, validate_default=True
    )
    """Dictionary of PyTorch training dataloader parameters. The dataloader parameters,
    should include the `shuffle` key, which is set to `True` by default. We strongly
    recommend to keep it as `True` to ensure the best training results."""

    val_dataloader_params: dict[str, Any] = Field(default={})
    """Dictionary of PyTorch validation dataloader parameters."""

    pred_dataloader_params: dict[str, Any] = Field(default={})
    """Dictionary of PyTorch prediction dataloader parameters."""

    seed: int | None = Field(default_factory=generate_random_seed, gt=0)
    """Random seed for reproducibility. If not specified, a random seed is generated."""

    @field_validator("axes")
    @classmethod
    def axes_valid(cls, axes: str, info: ValidationInfo) -> str:
        """
        Validate axes.

        Axes must:
        - be a combination of 'STCZYX'
        - not contain duplicates
        - contain at least 2 contiguous axes: X and Y
        - contain at most 4 axes
        - not contain both S and T axes

        Parameters
        ----------
        axes : str
            Axes to validate.
        info : ValidationInfo
            Validation information.

        Returns
        -------
        str
            Validated axes.

        Raises
        ------
        ValueError
            If axes are not valid.
        """
        # Additional validation for CZI files
        if info.data["data_type"] == "czi":
            if not check_czi_axes_validity(axes):
                raise ValueError(
                    f"Provided axes '{axes}' are not valid. Axes must be in the "
                    f"`SC(Z/T)YX` format, where Z or T are optional, and S and C can be"
                    f" singleton dimensions, but must be provided."
                )
        else:
            check_axes_validity(axes)

        return axes

    @field_validator("in_memory")
    @classmethod
    def validate_in_memory_with_data_type(cls, in_memory: bool, info: Any) -> bool:
        """
        Validate that in_memory is compatible with data_type.

        `in_memory` can only be True for 'array', 'tiff' and 'custom' data types.

        Parameters
        ----------
        in_memory : bool
            Whether to load data into memory.
        info : Any
            Additional information about the field being validated.

        Returns
        -------
        bool
            Validated in_memory value.

        Raises
        ------
        ValueError
            If in_memory is True for unsupported data types.
        """
        data_type = info.data.get("data_type")

        if in_memory and data_type not in ("array", "tiff", "custom"):
            raise ValueError(
                f"`in_memory` can only be True for 'array', 'tiff' and 'custom' "
                f"data types, got '{data_type}'. In memory loading of zarr and czi "
                f"data types is not currently not implemented."
            )

        if not in_memory and data_type == "array":
            raise ValueError(
                "`in_memory` must be True for 'array' data type, got False."
            )

        return in_memory

    @field_validator("channels", mode="before")
    @classmethod
    def validate_channels(
        cls,
        channels: Sequence[int] | None,
        info: ValidationInfo,
    ) -> Sequence[int] | None:
        """
        Validate channels.

        Channels must be a sequence of non-negative integers without duplicates. If
        channels are not `None`, then `C` must be present in the axes.

        Parameters
        ----------
        channels : Sequence of int or None
            Channels to validate.
        info : ValidationInfo
            Validation information.

        Returns
        -------
        Sequence of int or None
            Validated channels.

        Raises
        ------
        ValueError
            If channels are not valid.
        """
        if channels is not None:
            if "C" not in info.data["axes"]:
                raise ValueError(
                    "Channels were specified but 'C' is not present in the axes."
                )

            if isinstance(channels, int):
                channels = [channels]

            if not isinstance(channels, Sequence):
                raise ValueError("Channels must be a sequence of integers.")

            if len(channels) == 0:
                return None

            if not all(isinstance(ch, int) for ch in channels):
                raise ValueError("Channels must be integers.")

            if any(ch < 0 for ch in channels):
                raise ValueError("Channels must be non-negative integers.")

            if len(set(channels)) != len(channels):
                raise ValueError("Channels must not contain duplicates.")
        return channels

    @field_validator("patching")
    @classmethod
    def validate_patching_strategy_against_mode(
        cls, patching: PatchingConfig, info: ValidationInfo
    ) -> PatchingConfig:
        """
        Validate that the patching strategy is compatible with the dataset mode.

        - If mode is `training`, patching strategy must be `random`.
        - If mode is `validating`, patching must be `fixed_random`.
        - If mode is `predicting`, patching strategy must be `tiled` or `whole`.

        Parameters
        ----------
        patching : PatchingStrategies
            Patching strategy to validate.
        info : ValidationInfo
            Validation information.

        Returns
        -------
        PatchingStrategies
            Validated patching strategy.

        Raises
        ------
        ValueError
            If the patching strategy is not compatible with the dataset mode.
        """
        mode = info.data["mode"]
        if mode == Mode.TRAINING:
            if patching.name not in ["random", "stratified"]:
                # TODO: update error message for stratified patching
                raise ValueError(
                    f"Patching strategy '{patching.name}' is not compatible with "
                    f"mode '{mode.value}'. Use 'random' for training."
                )
        elif mode == Mode.VALIDATING:
            if patching.name != "fixed_random":
                raise ValueError(
                    f"Patching strategy '{patching.name}' is not compatible with "
                    f"mode '{mode.value}'. Use 'fixed_random' for validating."
                )
        elif mode == Mode.PREDICTING:
            if patching.name not in ["tiled", "whole"]:
                raise ValueError(
                    f"Patching strategy '{patching.name}' is not compatible with "
                    f"mode '{mode.value}'. Use 'tiled' or 'whole' for predicting."
                )
        return patching

    @field_validator("patch_filter", "coord_filter")
    @classmethod
    def validate_filters_against_mode(
        cls,
        filter_obj: PatchFilterConfig | CoordFilterConfig | None,
        info: ValidationInfo,
    ) -> PatchFilterConfig | CoordFilterConfig | None:
        """
        Validate that the filters are only used during training.

        Parameters
        ----------
        filter_obj : PatchFilters or CoordFilters or None
            Filter to validate.
        info : ValidationInfo
            Validation information.

        Returns
        -------
        PatchFilters or CoordFilters or None
            Validated filter.

        Raises
        ------
        ValueError
            If a filter is used in a mode other than training.
        """
        mode = info.data["mode"]
        if filter_obj is not None and mode != Mode.TRAINING:
            raise ValueError(
                f"Filter '{filter_obj.name}' can only be used in 'training' mode, "
                f"got mode '{mode.value}'."
            )
        return filter_obj

    @field_validator("train_dataloader_params")
    @classmethod
    def shuffle_train_dataloader(
        cls, train_dataloader_params: dict[str, Any]
    ) -> dict[str, Any]:
        """
        Validate that "shuffle" is included in the training dataloader params.

        A warning will be raised if `shuffle=False`.

        Parameters
        ----------
        train_dataloader_params : dict of {str: Any}
            The training dataloader parameters.

        Returns
        -------
        dict of {str: Any}
            The validated training dataloader parameters.

        Raises
        ------
        ValueError
            If "shuffle" is not included in the training dataloader params.
        """
        if "shuffle" not in train_dataloader_params:
            raise ValueError(
                "Value for 'shuffle' was not included in the `train_dataloader_params`."
            )
        elif ("shuffle" in train_dataloader_params) and (
            not train_dataloader_params["shuffle"]
        ):
            warn(
                "Dataloader parameters include `shuffle=False`, this will be passed to "
                "the training dataloader and may lead to lower quality results.",
                stacklevel=1,
            )
        return train_dataloader_params

    @model_validator(mode="after")
    def validate_dimensions(self: Self) -> Self:
        """
        Validate 2D/3D dimensions between axes and patch size.

        Returns
        -------
        Self
            Validated data model.

        Raises
        ------
        ValueError
            If the patch size dimension is not compatible with the axes.
        """
        # "whole" patching does not have dimensions to validate
        if not hasattr(self.patching, "patch_size"):
            return self

        if self.data_type == "czi":
            # Z and T are both depth axes for CZI data
            expected_dims = 3 if ("Z" in self.axes or "T" in self.axes) else 2
            additional_message = " (`Z` and `T` are depth axes for CZI data)"
        else:
            expected_dims = 3 if "Z" in self.axes else 2
            additional_message = ""

        # infer dimension from requested patch size
        actual_dims = len(self.patching.patch_size)
        if actual_dims != expected_dims:
            raise ValueError(
                f"`patch_size` in `patching` must have {expected_dims} dimensions, "
                f"got {self.patching.patch_size} with axes {self.axes}"
                f"{additional_message}."
            )

        return self

    @model_validator(mode="after")
    def propagate_seed_to_filters(self: Self) -> Self:
        """
        Propagate the main seed to patch and coordinate filters that support seeds.

        This ensures that all filters use the same seed for reproducibility,
        unless they already have a seed explicitly set.

        Returns
        -------
        Self
            Data model with propagated seeds.
        """
        if self.seed is not None:
            if self.patch_filter is not None:
                if (
                    hasattr(self.patch_filter, "seed")
                    and self.patch_filter.seed is None
                ):
                    self.patch_filter.seed = self.seed

            if self.coord_filter is not None:
                if (
                    hasattr(self.coord_filter, "seed")
                    and self.coord_filter.seed is None
                ):
                    self.coord_filter.seed = self.seed

        return self

    @model_validator(mode="after")
    def propagate_seed_to_transforms(self: Self) -> Self:
        """
        Propagate the main seed to all transforms that support seeds.

        This ensures that all transforms use the same seed for reproducibility,
        unless they already have a seed explicitly set.

        Returns
        -------
        Self
            Data model with propagated seeds.
        """
        if self.seed is not None:
            for transform in self.transforms:
                if hasattr(transform, "seed") and transform.seed is None:
                    transform.seed = self.seed
        return self

    @model_validator(mode="after")
    def propagate_seed_to_patching(self: Self) -> Self:
        """
        Propagate the main seed to the patching strategy if it supports seeds.

        This ensures that the patching strategy uses the same seed for reproducibility,
        unless it already has a seed explicitly set.

        Returns
        -------
        Self
            Data model with propagated seed.
        """
        if self.seed is not None:
            if hasattr(self.patching, "seed") and self.patching.seed is None:
                self.patching.seed = self.seed
        return self

    @field_validator("train_dataloader_params", "val_dataloader_params", mode="before")
    @classmethod
    def set_default_pin_memory(
        cls, dataloader_params: dict[str, Any]
    ) -> dict[str, Any]:
        """
        Set default pin_memory for dataloader parameters if not provided.

        - If 'pin_memory' is not set, it defaults to True if CUDA is available.

        Parameters
        ----------
        dataloader_params : dict of {str: Any}
            The dataloader parameters.

        Returns
        -------
        dict of {str: Any}
            The dataloader parameters with pin_memory default applied.
        """
        if "pin_memory" not in dataloader_params:
            import torch

            dataloader_params["pin_memory"] = torch.cuda.is_available()
        return dataloader_params

    @field_validator("train_dataloader_params", mode="before")
    @classmethod
    def set_default_train_workers(
        cls, dataloader_params: dict[str, Any]
    ) -> dict[str, Any]:
        """
        Set default num_workers for training dataloader if not provided.

        - If 'num_workers' is not set, it defaults to the number of available CPU cores.

        Parameters
        ----------
        dataloader_params : dict of {str: Any}
            The training dataloader parameters.

        Returns
        -------
        dict of {str: Any}
            The dataloader parameters with num_workers default applied.
        """
        if "num_workers" not in dataloader_params:
            # Use 0 workers during tests, otherwise use all available CPU cores
            if "pytest" in sys.modules:
                dataloader_params["num_workers"] = 0
            else:
                dataloader_params["num_workers"] = os.cpu_count()

        return dataloader_params

    @model_validator(mode="after")
    def set_val_workers_to_match_train(self: Self) -> Self:
        """
        Set validation dataloader num_workers to match training dataloader.

        If num_workers is not specified in val_dataloader_params, it will be set to the
        same value as train_dataloader_params["num_workers"].

        Returns
        -------
        Self
            Validated data model with synchronized num_workers.
        """
        if "num_workers" not in self.val_dataloader_params:
            self.val_dataloader_params["num_workers"] = self.train_dataloader_params[
                "num_workers"
            ]
        return self

    def __str__(self) -> str:
        """
        Pretty string reprensenting the configuration.

        Returns
        -------
        str
            Pretty string.
        """
        return pformat(self.model_dump())

    def _update(self, **kwargs: Any) -> None:
        """
        Update multiple arguments at once.

        Parameters
        ----------
        **kwargs : Any
            Keyword arguments to update.
        """
        self.__dict__.update(kwargs)
        self.__class__.model_validate(self.__dict__)

    def is_3D(self) -> bool:
        """
        Check if the data is 3D based on the axes.

        Either "Z" is in the axes and patching `patch_size` has 3 dimensions, or for CZI
        data, "Z" is in the axes or "T" is in the axes and patching `patch_size` has
        3 dimensions.

        This method is used during NGConfiguration validation to cross checks dimensions
        with the algorithm configuration.

        Returns
        -------
        bool
            True if the data is 3D, False otherwise.
        """
        if self.data_type == "czi":
            return "Z" in self.axes or "T" in self.axes
        else:
            return "Z" in self.axes

    # TODO: if switching from a state in which in_memory=True to an incompatible state
    # an error will be raised. Should that automatically be set to False instead?
    # TODO `channels=None` is ambigouous: all channels or same channels as in training?
    # TODO this method could be private and we could have public `to_validation_config`
    #   and `to_prediction_config` methods with appropriate parameters
    # TODO any use for switching to training mode?
    def convert_mode(
        self,
        new_mode: Literal["validating", "predicting"],
        new_patch_size: Sequence[int] | None = None,
        overlap_size: Sequence[int] | None = None,
        new_batch_size: int | None = None,
        new_data_type: Literal["array", "tiff", "zarr", "czi", "custom"] | None = None,
        new_axes: str | None = None,
        new_channels: Sequence[int] | Literal["all"] | None = None,
        new_in_memory: bool | None = None,
        new_dataloader_params: dict[str, Any] | None = None,
    ) -> NGDataConfig:
        """
        Convert a training dataset configuration to a different mode.

        This method is intended to facilitate creating validation or prediction
        configurations from a training configuration.

        To perform tile prediction when switching to `predicting` mode, please provide
        both `new_patch_size` and `overlap_size`. Switching mode to `predicting` without
        specifying `new_patch_size` and `overlap_size` will apply the default patching
        strategy, namely `whole` image strategy. `new_patch_size` and `overlap_size` are
        only used when switching to `predicting`.

        `channels=None` will retain the same channels as in the current configuration.
        To select all channels, please specify all channels explicitly or pass
        `channels='all'`.

        New dataloader parameters will be placed in the appropriate dataloader params
        field depending on the new mode.

        To create a new training configuration, please use
        `careamics.config.create_ng_data_configuration`.

        This method compares the new parameters with the current ones and raises
        errors if incompatible changes are requested, such as switching between 2D and
        3D axes, or changing the number of channels. Incompatibility across parameters
        may be delegated to Pydantic validation.

        Parameters
        ----------
        new_mode : Literal["validating", "predicting"]
            The new dataset mode, one of `validating` or `predicting`.
        new_patch_size : Sequence of int, default=None
            New patch size. If None for `predicting`, uses default whole image strategy.
        overlap_size : Sequence of int, default=None
            New overlap size. Necessary when switching to `predicting` with tiled
            patching.
        new_batch_size : int, default=None
            New batch size.
        new_data_type : Literal['array', 'tiff', 'zarr', 'czi', 'custom'], default=None
            New data type.
        new_axes : str, default=None
            New axes.
        new_channels : Sequence of int or "all", default=None
            New channels.
        new_in_memory : bool, default=None
            New in_memory value.
        new_dataloader_params : dict of {str: Any}, default=None
            New dataloader parameters. These will be placed in the
            appropriate dataloader params field depending on the new mode.

        Returns
        -------
        NGDataConfig
            New NGDataConfig with the updated mode and parameters.

        Raises
        ------
        ValueError
            If conversion to training mode is requested, or if incompatible changes
            are requested.
        """
        if new_mode == Mode.TRAINING:
            raise ValueError(
                "Converting to 'training' mode is not supported. Create a new "
                "NGDataConfig instead, for instance using "
                "`create_ng_data_configuration`."
            )
        if self.mode != Mode.TRAINING:
            raise ValueError(
                f"Converting from mode '{self.mode}' to '{new_mode}' is not supported. "
                f"Only conversion from 'training' mode is supported."
            )

        # sanity checks
        # switching spatial axes
        if new_axes is not None and ("Z" in new_axes) != (
            "Z" in self.axes
        ):  # switching 2D/3D
            raise ValueError("Cannot switch between 2D and 3D axes.")

        # normalize new_channels parameter to lift ambiguity around `None`
        #   - If None, keep previous parameter
        #   - If "all", select all channels (None value internally)
        if new_channels is None:
            new_channels = self.channels
        elif new_channels == "all":
            new_channels = None  # all channels

        # switching channels
        # if switching C axis:
        # - removing C: original channels can be `None`, singleton or multiple. New
        #   channels can be `None` if original were `None` or singleton, but not
        #   multiple.
        # - adding C: original channels can only be `None`. New channels can be `None`
        #   (but we warn users that they need to have a singleton C axis in the data),
        #   or singleton, but not multiple.
        adding_C_axis = (
            new_axes is not None and ("C" in new_axes) and ("C" not in self.axes)
        )
        removing_C_axis = (
            new_axes is not None and ("C" not in new_axes) and ("C" in self.axes)
        )
        prev_channels_not_singleton = self.channels is not None and (
            len(self.channels) != 1
        )

        if adding_C_axis:
            if new_channels is None:
                warn(
                    f"When switching to axes with 'C' (got {new_axes}) from axes "
                    f"{self.axes}, errors may be raised or degraded performances may be"
                    f" observed if the channel dimension in the data is not a singleton"
                    f" dimension. To select a specific channel, use the `new_channels` "
                    f"parameter (e.g. `new_channels=[1]`).",
                    stacklevel=1,
                )
            elif len(new_channels) != 1:
                raise ValueError(
                    f"When switching to axes with 'C' (got {new_axes}) from axes "
                    f"{self.axes}, a single channel only must be selected using the "
                    f"`new_channels` parameter (got {new_channels})."
                )
        elif removing_C_axis and prev_channels_not_singleton:
            raise ValueError(
                f"Cannot switch to axes without 'C' (got {new_axes}) from axes "
                f"{self.axes} when multiple channels were originally specified "
                f"({self.channels})."
            )

        # different number of channels
        if new_channels is not None and self.channels is not None:
            if len(new_channels) != len(self.channels):
                raise ValueError(
                    f"New channels length ({len(new_channels)}) does not match "
                    f"current channels length ({len(self.channels)})."
                )

        if self.channels is None and new_channels is not None:
            warn(
                f"Switching from all channels (`channels=None`) to specifying channels "
                f"{new_channels} may lead to errors or degraded performances if "
                f"{new_channels} are not all channels.",
                stacklevel=1,
            )  # Note that in the opposite case, self.channels is kept because
            # new_channels is None

        # apply default values
        patching_strategy: PatchingConfig
        if new_mode == Mode.PREDICTING:
            if new_patch_size is None:
                patching_strategy = WholePatchingConfig()
            else:
                if overlap_size is None:
                    raise ValueError(
                        "When switching to 'predicting' mode with 'tiled' patching, "
                        "the `overlap_size` parameter must be specified."
                    )
                patching_strategy = TiledPatchingConfig(
                    patch_size=list(new_patch_size), overlaps=list(overlap_size)
                )
        else:  # validating
            assert not isinstance(self.patching, WholePatchingConfig)

            patching_strategy = FixedRandomPatchingConfig(
                patch_size=(
                    list(new_patch_size)
                    if new_patch_size is not None
                    else self.patching.patch_size
                ),
            )

        # create new config
        model_dict = self.model_dump()
        model_dict.update(
            {
                "mode": new_mode,
                "patching": patching_strategy,
                "batch_size": new_batch_size or self.batch_size,
                "data_type": new_data_type or self.data_type,
                "axes": new_axes or self.axes,
                "channels": new_channels if new_channels is not None else self.channels,
                "in_memory": (
                    new_in_memory if new_in_memory is not None else self.in_memory
                ),
                "val_dataloader_params": (
                    new_dataloader_params
                    if new_mode == Mode.VALIDATING and new_dataloader_params is not None
                    else self.val_dataloader_params
                ),
                "pred_dataloader_params": (
                    new_dataloader_params
                    if new_mode == Mode.PREDICTING and new_dataloader_params is not None
                    else self.pred_dataloader_params
                ),
            }
        )

        # remove patch and coord filters when switching to validation or prediction
        del model_dict["patch_filter"]
        del model_dict["coord_filter"]

        return NGDataConfig(**model_dict)

    # def set_3D(self, axes: str, patch_size: list[int]) -> None:
    #     """
    #     Set 3D parameters.

    #     Parameters
    #     ----------
    #     axes : str
    #         Axes.
    #     patch_size : list of int
    #         Patch size.
    #     """
    #     self._update(axes=axes, patch_size=patch_size)



"""Pydantic model for the mask coordinate filter."""

from typing import Literal

from pydantic import Field

from .filter_config import FilterConfig


class MaskFilterConfig(FilterConfig):
    """Pydantic model for the mask coordinate filter."""

    name: Literal["mask"] = "mask"
    """Name of the filter."""

    coverage: float = Field(0.5, ge=0.0, le=1.0)
    """Percentage of masked pixels required to keep a patch."""



"""CAREamics transformation Pydantic models."""

__all__ = [
    "NORM_AND_SPATIAL_UNION",
    "SPATIAL_TRANSFORMS_UNION",
    "N2VManipulateConfig",
    "NormalizeConfig",
    "TransformConfig",
    "XYFlipConfig",
    "XYRandomRotate90Config",
]


from .n2v_manipulate_config import N2VManipulateConfig
from .normalize_config import NormalizeConfig
from .transform_config import TransformConfig
from .transform_unions import (
    NORM_AND_SPATIAL_UNION,
    SPATIAL_TRANSFORMS_UNION,
)
from .xy_flip_config import XYFlipConfig
from .xy_random_rotate90_config import XYRandomRotate90Config



"""Factories for coordinate and patch filters."""

from typing import Union

from careamics.config.data.patch_filter import (
    FilterConfig,
    MaskFilterConfig,
    MaxFilterConfig,
    MeanSTDFilterConfig,
    ShannonFilterConfig,
)
from careamics.config.support.supported_filters import (
    SupportedCoordinateFilters,
    SupportedPatchFilters,
)
from careamics.dataset_ng.image_stack import GenericImageStack
from careamics.dataset_ng.patch_extractor import PatchExtractor

from .mask_filter import MaskCoordFilter
from .max_filter import MaxPatchFilter
from .mean_std_filter import MeanStdPatchFilter
from .shannon_filter import ShannonPatchFilter

PatchFilter = Union[
    MaxPatchFilter,
    MeanStdPatchFilter,
    ShannonPatchFilter,
]


CoordFilter = Union[MaskCoordFilter]


def create_coord_filter(
    filter_model: FilterConfig, mask: PatchExtractor[GenericImageStack]
) -> CoordFilter:
    """Factory function to create coordinate filter instances based on the filter name.

    Parameters
    ----------
    filter_model : FilterModel
        Pydantic model of the filter to be created.
    mask : PatchExtractor[GenericImageStack]
        Mask extractor to be used for the mask filter.

    Returns
    -------
    CoordFilter
        Instance of the mask patch filter.
    """
    if filter_model.name == SupportedCoordinateFilters.MASK:
        assert isinstance(filter_model, MaskFilterConfig)
        return MaskCoordFilter(
            mask_extractor=mask,
            coverage=filter_model.coverage,
            p=filter_model.p,
            seed=filter_model.seed,
        )
    else:
        raise ValueError(f"Unknown filter name: {filter_model}")


def create_patch_filter(filter_model: FilterConfig) -> PatchFilter:
    """Factory function to create patch filter instances based on the filter name.

    Parameters
    ----------
    filter_model : FilterModel
        Pydantic model of the filter to be created.

    Returns
    -------
    PatchFilter
        Instance of the requested patch filter.
    """
    if filter_model.name == SupportedPatchFilters.MAX:
        assert isinstance(filter_model, MaxFilterConfig)
        return MaxPatchFilter(
            threshold=filter_model.threshold, p=filter_model.p, seed=filter_model.seed
        )
    elif filter_model.name == SupportedPatchFilters.MEANSTD:
        assert isinstance(filter_model, MeanSTDFilterConfig)
        return MeanStdPatchFilter(
            mean_threshold=filter_model.mean_threshold,
            std_threshold=filter_model.std_threshold,
            p=filter_model.p,
            seed=filter_model.seed,
        )
    elif filter_model.name == SupportedPatchFilters.SHANNON:
        assert isinstance(filter_model, ShannonFilterConfig)
        return ShannonPatchFilter(
            threshold=filter_model.threshold, p=filter_model.p, seed=filter_model.seed
        )
    else:
        raise ValueError(f"Unknown filter name: {filter_model}")



import random
from pathlib import Path

import numpy as np
import pytest
from numpy.typing import NDArray
from tifffile import imwrite

from careamics.config.data import NGDataConfig
from careamics.dataset.dataset_utils import reshape_array
from careamics.dataset_ng.dataset import CareamicsDataset, ImageRegionData
from careamics.dataset_ng.image_stack_loader import load_arrays, load_tiffs
from careamics.dataset_ng.patch_extractor import PatchExtractor
from careamics.dataset_ng.patching_strategies import TilingStrategy
from careamics.lightning.dataset_ng.prediction.stitch_prediction import (
    group_tiles_by_key,
    stitch_prediction,
    stitch_single_prediction,
    stitch_single_sample,
)


@pytest.fixture
def data_config(data_type, axes, shape, channels) -> NGDataConfig:
    # create tiling strategy
    if "Z" in axes:
        tile_size = (8, 16, 16)
        overlaps = (2, 4, 4)
    else:
        tile_size = (16, 16)
        overlaps = (4, 4)

    if "C" in axes:
        n_channels = shape[axes.index("C")]

        if channels is not None:
            n_channels = len(channels)
    else:
        n_channels = 1

    return NGDataConfig(
        mode="predicting",
        data_type=data_type,
        patching={
            "name": "tiled",
            "patch_size": tile_size,
            "overlaps": overlaps,
        },
        axes=axes,
        channels=channels,
        normalization={
            "name": "mean_std",
            "input_means": [0.0 for _ in range(n_channels)],
            "input_stds": [1.0 for _ in range(n_channels)],
        },
    )


@pytest.fixture
def tiles(
    tmp_path, data_config: NGDataConfig, n_data, shape
) -> tuple[NDArray, list[ImageRegionData]]:
    """Create tiles.

    Note that the tiles will be slightly different because the dataset changes the dtype
    of the data and performs normalization.

    We must use np.testing.assert_allclose(stitched_array, array, rtol=1e-5, atol=0),
    with relative tolerance as the errors scale with the values and we use
    np.arange to create the data.

    Returns
    -------
    np.ndarray
        Original array of shape DSC(Z)YX, where D is data.
    list of ImageRegionData
        Extracted tiles.
    """
    # create data
    array = np.arange(n_data * np.prod(shape)).reshape((n_data, *shape))

    if data_config.data_type == "tiff":
        sources = []
        root = tmp_path / "tiff_data"
        root.mkdir(parents=True, exist_ok=True)

        for i in range(n_data):
            file_path = root / f"array_{i}.tiff"
            sources.append(file_path)

            write_array = array[i]

            imwrite(file_path, write_array)
    else:  # array
        sources = [array[i] for i in range(n_data)]

    if "S" in data_config.axes:
        if "C" in data_config.axes:
            shape_with_sc = shape
        else:
            shape_with_sc = (shape[0], 1, *shape[1:])
    else:
        if "C" in data_config.axes:
            shape_with_sc = (1, *shape)
        else:
            shape_with_sc = (1, 1, *shape)

    tiling_strategy = TilingStrategy(
        data_shapes=[shape_with_sc] * n_data,
        patch_size=data_config.patching.patch_size,
        overlaps=data_config.patching.overlaps,
    )
    n_tiles = tiling_strategy.n_patches

    # create patch extractor
    if data_config.data_type == "tiff":
        image_stacks = load_tiffs(source=sources, axes=data_config.axes)
    else:
        image_stacks = load_arrays(source=sources, axes=data_config.axes)
    patch_extractor = PatchExtractor(image_stacks)

    # create dataset
    dataset = CareamicsDataset(
        data_config=data_config,
        input_extractor=patch_extractor,
    )

    # extract tiles
    tiles: list[ImageRegionData] = []
    for i in range(n_tiles):
        tiles.append(dataset[i][0])

    # reshape array for testing
    arrays = [reshape_array(array[i], data_config.axes) for i in range(array.shape[0])]

    return np.stack(arrays, axis=0), tiles


# parameters are injected automatically in the tiles fixture
@pytest.mark.parametrize("n_data", [1, 3])
@pytest.mark.parametrize(
    "data_type, shape, axes, channels",
    [
        ("array", (32, 32), "YX", None),
    ],
)
def test_group_tiles_by_data_index(tiles):
    array, tile_list = tiles
    n_data = array.shape[0]

    sorted_tiles = group_tiles_by_key(tile_list, key="data_idx")
    assert len(sorted_tiles.keys()) == n_data
    n_tiles_per_data = [len(sorted_tiles[data_idx]) for data_idx in sorted_tiles.keys()]
    assert all(n_tiles_per_data[0] == n_tiles_per_data[i] for i in range(1, n_data))


# parameters are injected automatically in the tiles fixture
@pytest.mark.parametrize(
    "n_data, data_type, shape, axes, channels",
    [
        (1, "array", (5, 32, 32), "SYX", None),
    ],
)
def test_group_tiles_by_sample_index(tiles):
    array, tile_list = tiles
    n_data = array.shape[1]

    sorted_tiles = group_tiles_by_key(tile_list, key="sample_idx")
    assert len(sorted_tiles.keys()) == n_data
    n_tiles_per_data = [len(sorted_tiles[data_idx]) for data_idx in sorted_tiles.keys()]
    assert all(n_tiles_per_data[0] == n_tiles_per_data[i] for i in range(1, n_data))


# parameters are injected automatically in the tiles fixture
@pytest.mark.parametrize("n_data", [1])
@pytest.mark.parametrize(
    "data_type, shape, axes, channels",
    [
        ("array", (32, 32), "YX", None),
        ("array", (3, 32, 32), "CYX", None),
        ("array", (3, 32, 32), "CYX", [1]),
        ("array", (3, 32, 32), "CYX", [0, 2]),
        ("array", (8, 32, 32), "ZYX", None),
        ("array", (3, 8, 32, 32), "CZYX", None),
        ("array", (3, 8, 32, 32), "CZYX", [1]),
        ("array", (3, 8, 32, 32), "CZYX", [0, 2]),
    ],
)
def test_stitching_single_sample(tiles, channels):
    array, tile_list = tiles

    # n_data = 1, so remove data dim
    # no sample dim present, so remove it as well
    array = array.squeeze(axis=0).squeeze(axis=0)

    # adjust channels
    if channels is not None:
        array = array[channels]

    stitched_array = stitch_single_sample(tile_list)
    np.testing.assert_allclose(stitched_array, array, rtol=1e-5, atol=0)


# parameters are injected automatically in the tiles fixture
@pytest.mark.parametrize("n_data", [1])
@pytest.mark.parametrize(
    "data_type, shape, axes, channels",
    [
        ("array", (32, 32), "YX", None),
        ("array", (3, 32, 32), "CYX", None),
        ("array", (3, 32, 32), "CYX", [1]),
        ("array", (3, 32, 32), "CYX", [0, 2]),
        ("array", (5, 32, 32), "SYX", None),
        ("array", (8, 32, 32), "ZYX", None),
        ("array", (3, 8, 32, 32), "CZYX", None),
        ("array", (3, 8, 32, 32), "CZYX", [1]),
        ("array", (3, 8, 32, 32), "CZYX", [0, 2]),
        ("array", (5, 3, 32, 32), "SCYX", None),
        ("array", (5, 3, 32, 32), "SCYX", [1]),
        ("array", (5, 3, 32, 32), "SCYX", [0, 2]),
        ("array", (5, 8, 32, 32), "SZYX", None),
        ("array", (5, 3, 8, 32, 32), "SCZYX", None),
        ("array", (5, 3, 8, 32, 32), "SCZYX", [1]),
        ("array", (5, 3, 8, 32, 32), "SCZYX", [0, 2]),
    ],
)
def test_stitching_single_prediction(tiles, channels):
    array, tile_list = tiles

    # n_data = 1, so remove data dim
    array = array.squeeze(axis=0)

    # adjust channels
    if channels is not None:
        array = array[:, channels]

    stitched_array = stitch_single_prediction(tile_list)
    np.testing.assert_allclose(stitched_array, array, rtol=1e-5, atol=0)


# parameters are injected automatically in the tiles fixture
@pytest.mark.parametrize("n_data", [4])
@pytest.mark.parametrize(
    "data_type, shape, axes, channels",
    [
        ("tiff", (32, 32), "YX", None),
        ("tiff", (3, 32, 32), "CYX", None),
        ("tiff", (3, 32, 32), "CYX", [1]),
        ("tiff", (3, 32, 32), "CYX", [0, 2]),
        ("tiff", (5, 32, 32), "SYX", None),
        ("tiff", (8, 32, 32), "ZYX", None),
        ("tiff", (3, 8, 32, 32), "CZYX", None),
        ("tiff", (3, 8, 32, 32), "CZYX", [1]),
        ("tiff", (3, 8, 32, 32), "CZYX", [0, 2]),
        ("tiff", (5, 3, 32, 32), "SCYX", None),
        ("tiff", (5, 3, 32, 32), "SCYX", [1]),
        ("tiff", (5, 3, 32, 32), "SCYX", [0, 2]),
        ("tiff", (5, 8, 32, 32), "SZYX", None),
        ("tiff", (5, 3, 8, 32, 32), "SCZYX", None),
        ("tiff", (5, 3, 8, 32, 32), "SCZYX", [1]),
        ("tiff", (5, 3, 8, 32, 32), "SCZYX", [0, 2]),
    ],
)
def test_stitching_prediction(tiles, channels):
    array, tile_list = tiles

    stitched_arrays, data_idx = stitch_prediction(tile_list)

    # adjust for channels
    if channels is not None:
        array = array[:, :, channels]

    prediction = np.stack(stitched_arrays)
    np.testing.assert_allclose(prediction, array, rtol=1e-5, atol=0)

    # test data indices
    for i, data_idx_str in enumerate(data_idx):
        assert Path(data_idx_str).name == f"array_{i}.tiff"


# parameters are injected automatically in the tiles fixture
@pytest.mark.parametrize("n_data", [4])
@pytest.mark.parametrize(
    "data_type, shape, axes, channels",
    [
        ("tiff", (32, 32), "YX", None),
    ],
)
def test_stitching_prediction_ordering(tiles, channels):
    array, tile_list = tiles

    # shuffle tiles to test ordering
    random.shuffle(tile_list)

    stitched_arrays, sources = stitch_prediction(tile_list)

    # check that data indices are in order
    for i, data_idx_str in enumerate(sources):
        assert Path(data_idx_str).name == f"array_{i}.tiff"
        np.testing.assert_allclose(stitched_arrays[i], array[i], rtol=1e-5, atol=0)



__all__ = [
    "FixedRandomPatchingStrategy",
    "PatchSpecs",
    "PatchingStrategy",
    "RandomPatchingStrategy",
    "RegionSpecs",
    "SequentialPatchingStrategy",
    "StratifiedPatchingStrategy",
    "TileSpecs",
    "TilingStrategy",
    "WholeSamplePatchingStrategy",
    "create_patching_strategy",
    "is_tile_specs",
]

from .patching_strategy_factory import create_patching_strategy
from .patching_strategy_protocol import (
    PatchingStrategy,
    PatchSpecs,
    RegionSpecs,
    TileSpecs,
    is_tile_specs,
)
from .random_patching import FixedRandomPatchingStrategy, RandomPatchingStrategy
from .sequential_patching import SequentialPatchingStrategy
from .stratified_patching import StratifiedPatchingStrategy
from .tiling_strategy import TilingStrategy
from .whole_sample import WholeSamplePatchingStrategy



"""HDN algorithm configuration."""

from typing import Literal

from bioimageio.spec.generic.v0_3 import CiteEntry
from pydantic import ConfigDict

from careamics.config.algorithms.vae_algorithm_config import VAEBasedAlgorithm
from careamics.config.architectures import LVAEConfig
from careamics.config.losses.loss_config import LVAELossConfig

HDN = "Hierarchical DivNoising"

HDN_DESCRIPTION = (
    "HDN leverages a hierarchical VAE to perform image "
    "restoration. It is designed to be interpretable and unsupervised, "
    "making it suitable for a wide range of microscopy images."
)
HDN_REF = CiteEntry(
    text='Prakash, M., Delbracio, M., Milanfar, P., Jug, F. 2022. "Interpretable '
    'Unsupervised Diversity Denoising and Artefact Removal." The International '
    "Conference on Learning Representations (ICLR).",
    doi="10.1561/2200000056",
)


class HDNAlgorithm(VAEBasedAlgorithm):
    """HDN algorithm configuration."""

    model_config = ConfigDict(validate_assignment=True)

    algorithm: Literal["hdn"] = "hdn"

    loss: LVAELossConfig

    model: LVAEConfig  # TODO add validators

    is_supervised: bool = False

    def get_algorithm_friendly_name(self) -> str:
        """
        Get the algorithm friendly name.

        Returns
        -------
        str
            Friendly name of the algorithm.
        """
        return HDN

    def get_algorithm_keywords(self) -> list[str]:
        """
        Get algorithm keywords.

        Returns
        -------
        list[str]
            List of keywords.
        """
        return [
            "restoration",
            "VAE",
            "3D" if self.model.is_3D() else "2D",
            "CAREamics",
            "pytorch",
        ]

    def get_algorithm_references(self) -> str:
        """
        Get the algorithm references.

        This is used to generate the README of the BioImage Model Zoo export.

        Returns
        -------
        str
            Algorithm references.
        """
        return HDN_REF.text + " doi: " + HDN_REF.doi

    def get_algorithm_citations(self) -> list[CiteEntry]:
        """
        Return a list of citation entries of the current algorithm.

        This is used to generate the model description for the BioImage Model Zoo.

        Returns
        -------
        List[CiteEntry]
            List of citation entries.
        """
        return [HDN_REF]

    def get_algorithm_description(self) -> str:
        """
        Get the algorithm description.

        Returns
        -------
        str
            Algorithm description.
        """
        return HDN_DESCRIPTION



import numpy as np
from skimage.measure import shannon_entropy

from careamics.dataset_ng.patch_filter import ShannonPatchFilter


def test_shannon_filter():
    """Test ShannonEntropyFilter functionality."""
    rng = np.random.default_rng(42)

    size = 8
    img = np.zeros((size, size))
    img[size // 2 :, size // 2 :] = rng.integers(100, 255, (size // 2, size // 2))

    corner_entropy = shannon_entropy(img[size // 2 :, size // 2 :])

    shannon_filter = ShannonPatchFilter(threshold=corner_entropy / 2)

    # exclude border
    assert shannon_filter.filter_out(img[0:4, 0:4])  # corner, no entropy

    # exclude low entropy
    assert shannon_filter.filter_out(img[2:6, 2:6])  # quarter bg, low entropy

    # include higher entropy
    assert not shannon_filter.filter_out(img[2:6, 4:])  # half bg, higher entropy

    # include high entropy
    assert not shannon_filter.filter_out(img[4:, 4:])  # full fg, high entropy
